from __future__ import absolute_import, division, print_function

import uuid
from copy import copy, deepcopy
from datetime import datetime
from typing import TYPE_CHECKING
from urllib.parse import urlparse, urlunsplit

import attr

from . import deprecated
from .accessibility import AccessibilitySettings
from .geometry import RectangleSize
from .match import ImageMatchSettings, MatchLevel
from .server import FailureReports, SessionType
from .utils import argument_guard
from .utils.converters import str2bool
from .utils.datetime_utils import UTC
from .utils.general_utils import get_env_with_prefix

if TYPE_CHECKING:
    from typing import TYPE_CHECKING, Dict, List, Optional, Text, TypeVar

    from .utils.custom_types import ViewPort

    Self = TypeVar("Self", bound="Configuration")  # typedef

__all__ = ("BatchInfo", "Configuration")

MINIMUM_MATCH_TIMEOUT_MS = 600
DEFAULT_ALL_TEST_RESULTS_TIMEOUT = 30 * 60  # seconds
PROCESS_DEFAULT_BATCH_ID = str(uuid.uuid4())  # Unique per-process


@attr.s(init=False, slots=True)
class BatchInfo(object):
    """
    A batch of tests.
    """

    name = attr.ib()  # type: Text
    started_at = attr.ib()  # type: datetime
    sequence_name = attr.ib()  # type: Optional[Text]
    id = attr.ib()  # type: Text
    notify_on_completion = attr.ib()  # type: bool
    properties = attr.ib()  # type: List[Dict[Text,Text]]

    def __init__(self, name=None, started_at=None, batch_sequence_name=None):
        # type: (Optional[Text], Optional[datetime], Optional[Text]) -> None
        self.id = get_env_with_prefix(
            "APPLITOOLS_BATCH_ID", PROCESS_DEFAULT_BATCH_ID
        )  # type: Text
        self.name = get_env_with_prefix("APPLITOOLS_BATCH_NAME")  # type: Text
        self.started_at = datetime.now(UTC)  # type: datetime
        self.sequence_name = get_env_with_prefix(
            "APPLITOOLS_BATCH_SEQUENCE"
        )  # type: Optional[Text]
        self.notify_on_completion = str2bool(
            get_env_with_prefix("APPLITOOLS_BATCH_NOTIFY")
        )  # type: bool
        self.properties = []  # type: List[Dict[Text,Text]]

        if name:
            self.name = name
        if started_at:
            self.started_at = started_at
        if batch_sequence_name:
            self.sequence_name = batch_sequence_name

    def with_batch_id(self, id):
        # type: (Text) -> BatchInfo
        argument_guard.not_none(id)
        self.id = str(id)
        return self

    def add_property(self, name, value):
        # type: (Text, Text) -> BatchInfo
        """
        Associates a key/value pair with the Batch. This can be used later for filtering.
        :param name: (string) The property name.
        :param value: (string) The property value
        """
        self.properties.append({"name": name, "value": value})
        return self

    def clear_properties(self):
        # type: () -> BatchInfo
        """
        Clears the list of Batch properties.
        """
        del self.properties[:]
        return self


class ProxySettings(object):
    def __init__(
        self,
        host_or_url,  # type: Text
        port=None,  # type: Optional[int]
        username=None,  # type: Optional[Text]
        password=None,  # type: Optional[Text]
        scheme=None,  # type: Text
    ):
        # type: (...) -> None
        if host_or_url.startswith("http://") or host_or_url.startswith("https://"):
            parsed = urlparse(host_or_url)
            self.host = parsed.hostname
            self.port = port or parsed.port or 8888
            self.username = username or parsed.username
            self.password = password or parsed.password
            self.scheme = scheme or parsed.scheme or "http"
        else:
            self.host = host_or_url
            self.port = port or 8888
            self.username = username
            self.password = password
            self.scheme = scheme or "http"

    @property
    def url(self):
        password = ":{}".format(self.password) if self.password else ""
        auth = "{}{}@".format(self.username, password) if self.username else ""
        port = ":{}".format(self.port) if self.port else ""
        return urlunsplit((self.scheme, auth + self.host + port, "", "", ""))

    @classmethod
    def from_env(cls):
        # type: (ProxySettings) -> ProxySettings | None
        url = get_env_with_prefix("APPLITOOLS_HTTP_PROXY")
        if url:
            return cls(url)
        return None


@attr.s
class Configuration(object):
    batch = attr.ib(factory=BatchInfo)  # type: BatchInfo
    branch_name = attr.ib(
        factory=lambda: get_env_with_prefix("APPLITOOLS_BRANCH", None),
    )  # type: Optional[Text]
    parent_branch_name = attr.ib(
        factory=lambda: get_env_with_prefix("APPLITOOLS_PARENT_BRANCH", None),
    )  # type: Optional[Text]
    baseline_branch_name = attr.ib(
        factory=lambda: get_env_with_prefix("APPLITOOLS_BASELINE_BRANCH", None),
    )  # type: Optional[Text]
    agent_id = attr.ib(default=None)  # type: Optional[Text]
    baseline_env_name = attr.ib(default=None)  # type: Optional[Text]
    environment_name = attr.ib(default=None)  # type: Optional[Text]
    save_diffs = attr.ib(default=None)  # type: bool
    app_name = attr.ib(default=None)  # type: Optional[Text]
    test_name = attr.ib(default=None)  # type: Optional[Text]
    viewport_size = attr.ib(
        default=None,
        converter=attr.converters.optional(RectangleSize.from_),
    )  # type: Optional[RectangleSize]
    session_type = attr.ib(default=None)  # type: Optional[SessionType]
    host_app = attr.ib(default=None)  # type: Optional[Text]
    host_os = attr.ib(default=None)  # type: Optional[Text]
    properties = attr.ib(factory=list)  # type: List[Dict[Text, Text]]
    match_timeout = attr.ib(default=None)  # type: Optional[int] # ms
    is_disabled = attr.ib(default=None)  # type: Optional[bool]
    save_new_tests = attr.ib(default=None)  # type: Optional[bool]
    save_failed_tests = attr.ib(default=None)  # type: Optional[bool]
    failure_reports = attr.ib(default=None)  # type: Optional[FailureReports]
    send_dom = attr.ib(default=None)  # type: Optional[bool]
    default_match_settings = attr.ib(
        factory=ImageMatchSettings
    )  # type: ImageMatchSettings
    stitch_overlap = attr.ib(default=None)  # type: Optional[int]
    api_key = attr.ib(
        factory=lambda: get_env_with_prefix("APPLITOOLS_API_KEY", None)
    )  # type: Optional[Text]
    server_url = attr.ib(
        factory=lambda: get_env_with_prefix("APPLITOOLS_SERVER_URL"),
    )  # type: Text
    _timeout = attr.ib(default=None)  # type: Optional[int] # ms
    proxy = attr.ib(factory=ProxySettings.from_env)  # type: ProxySettings | None
    save_debug_screenshots = attr.ib(default=None)  # type: Optional[bool]
    debug_screenshots_path = attr.ib(
        converter=str, factory=lambda: get_env_with_prefix("DEBUG_SCREENSHOT_PATH", "")
    )
    debug_screenshots_prefix = attr.ib(
        converter=str,
        factory=lambda: get_env_with_prefix("DEBUG_SCREENSHOT_PREFIX", "screenshot_"),
    )
    wait_before_capture = attr.ib(default=None)  # type: Optional[int]
    user_test_id = attr.ib(default=None)  # type: Optional[Text]
    disable_nml_url_cache = attr.ib(default=None)  # type: Optional[bool]

    @property
    def enable_patterns(self):
        # type: () -> bool
        return self.default_match_settings.enable_patterns

    @enable_patterns.setter
    def enable_patterns(self, enable_patterns):
        # type: (bool) -> None
        self.default_match_settings.enable_patterns = enable_patterns

    @property
    def use_dom(self):
        # type: () -> bool
        return self.default_match_settings.use_dom

    @use_dom.setter
    def use_dom(self, use_dom):
        # type: (bool) -> None
        self.default_match_settings.use_dom = use_dom

    @property
    def match_level(self):
        # type: () -> MatchLevel
        return self.default_match_settings.match_level

    @match_level.setter
    def match_level(self, match_level):
        # type: (MatchLevel) -> None
        self.default_match_settings.match_level = match_level

    @property
    def ignore_displacements(self):
        # type: () -> bool
        return self.default_match_settings.ignore_displacements

    @ignore_displacements.setter
    def ignore_displacements(self, ignore_displacements):
        # type: (bool) -> None
        self.default_match_settings.ignore_displacements = ignore_displacements

    def set_batch(self, batch):
        # type: (Self, BatchInfo) -> Self
        argument_guard.is_a(batch, BatchInfo)
        self.batch = batch
        return self

    def set_branch_name(self, branch_name):
        # type: (Self, Text) -> Self
        self.branch_name = branch_name
        return self

    def set_agent_id(self, agent_id):
        # type: (Self, Text) -> Self
        self.agent_id = agent_id
        return self

    def set_parent_branch_name(self, parent_branch_name):
        # type: (Self, Text) -> Self
        self.parent_branch_name = parent_branch_name
        return self

    def set_baseline_branch_name(self, baseline_branch_name):
        # type: (Self, Text) -> Self
        self.baseline_branch_name = baseline_branch_name
        return self

    def set_baseline_env_name(self, baseline_env_name):
        # type: (Self, Text) -> Self
        self.baseline_env_name = baseline_env_name
        return self

    def set_environment_name(self, environment_name):
        # type: (Self, Text) -> Self
        self.environment_name = environment_name
        return self

    def set_save_diffs(self, save_diffs):
        # type: (Self, bool) -> Self
        self.save_diffs = save_diffs
        return self

    def set_app_name(self, app_name):
        # type: (Self, Text) -> Self
        self.app_name = app_name
        return self

    def set_test_name(self, test_name):
        # type: (Self, Text) -> Self
        self.test_name = test_name
        return self

    def set_viewport_size(self, viewport_size):
        # type: (Self, ViewPort) -> Self
        self.viewport_size = viewport_size
        return self

    def set_session_type(self, session_type):
        # type: (Self, SessionType) -> Self
        self.session_type = session_type
        return self

    @property
    def ignore_caret(self):
        # type: () -> bool
        return self.default_match_settings.ignore_caret

    def set_ignore_caret(self, ignore_caret):
        # type: (Self, bool) -> Self
        self.default_match_settings.ignore_caret = ignore_caret
        return self

    def set_host_app(self, host_app):
        # type: (Self, Text) -> Self
        self.host_app = host_app
        return self

    def set_host_os(self, host_os):
        # type: (Self, Text) -> Self
        self.host_os = host_os
        return self

    def set_match_timeout(self, match_timeout):
        # type: (Self, int) -> Self
        self.match_timeout = match_timeout
        return self

    def set_match_level(self, match_level):
        # type: (Self, MatchLevel) -> Self
        self.match_level = match_level
        return self

    def set_ignore_displacements(self, ignore_displacements):
        # type: (Self, bool) -> Self
        self.ignore_displacements = ignore_displacements
        return self

    def set_save_new_tests(self, save_new_tests):
        # type: (Self, bool) -> Self
        self.save_new_tests = save_new_tests
        return self

    def set_save_failed_tests(self, save_failed_tests):
        # type: (Self, bool) -> Self
        self.save_failed_tests = save_failed_tests
        return self

    def set_failure_reports(self, failure_reports):
        # type: (Self, FailureReports) -> Self
        self.failure_reports = failure_reports
        return self

    def set_send_dom(self, send_dom):
        # type: (Self, bool) -> Self
        self.send_dom = send_dom
        return self

    def set_use_dom(self, use_dom):
        # type: (Self, bool) -> Self
        self.use_dom = use_dom
        return self

    def set_enable_patterns(self, enable_patterns):
        # type: (Self, bool) -> Self
        self.enable_patterns = enable_patterns
        return self

    def set_stitch_overlap(self, stitch_overlap):
        # type: (Self, int) -> Self
        self.stitch_overlap = stitch_overlap
        return self

    def set_api_key(self, api_key):
        # type: (Self, Text) -> Self
        self.api_key = api_key
        return self

    def set_server_url(self, server_url):
        # type: (Self, Text) -> Self
        self.server_url = server_url
        return self

    def set_default_match_settings(self, default_match_settings):
        # type: (Self, ImageMatchSettings) -> Self
        self.default_match_settings = default_match_settings
        return self

    @property
    def accessibility_validation(self):
        # type: (Self) -> Optional[AccessibilitySettings]
        return self.default_match_settings.accessibility_settings

    @accessibility_validation.setter
    def accessibility_validation(self, accessibility_settings):
        # type: (Self, Optional[AccessibilitySettings]) -> None
        if accessibility_settings is None:
            self.self.default_match_settings.accessibility_settings = None
            return
        argument_guard.is_a(accessibility_settings, AccessibilitySettings)
        self.default_match_settings.accessibility_settings = accessibility_settings

    def set_accessibility_validation(self, accessibility_settings):
        # type: (Self, Optional[AccessibilitySettings]) -> Self
        self.accessibility_validation = accessibility_settings
        return self

    @match_timeout.validator
    def _validate1(self, attribute, value):
        if value is not None and 0 < value < MINIMUM_MATCH_TIMEOUT_MS:
            raise ValueError(
                "Match timeout must be at least {} ms.".format(MINIMUM_MATCH_TIMEOUT_MS)
            )

    @viewport_size.validator
    def _validate2(self, attribute, value):
        if value is None:
            return None
        if isinstance(value, RectangleSize) or (
            isinstance(value, dict)
            and "width" in value.keys()
            and "height" in value.keys()
        ):
            return None

        raise ValueError("Wrong viewport type settled")

    @property
    def is_send_dom(self):
        # type: () -> bool
        return self.send_dom

    def clone(self):
        # type: () -> Self
        # TODO: Remove this huck when get rid of Python2
        # deepcopy on python 2 raise an exception so construct manually
        conf = copy(self)
        conf.batch = deepcopy(conf.batch)
        conf.viewport_size = copy(conf.viewport_size)
        conf.properties = deepcopy(conf.properties)
        conf.default_match_settings = deepcopy(conf.default_match_settings)
        return conf

    def add_property(self, name, value):
        # type: (Text, Text) -> Self
        """
        Associates a key/value pair with the test. This can be used later for filtering.
        :param name: (string) The property name.
        :param value: (string) The property value
        """
        self.properties.append({"name": name, "value": value})
        return self

    def clear_properties(self):
        # type: () -> Self
        """
        Clears the list of custom properties.
        """
        del self.properties[:]
        return self

    def set_proxy(self, proxy):
        # type: (Optional[ProxySettings | Text]) -> Self
        if isinstance(proxy, ProxySettings):
            self.proxy = proxy
        elif isinstance(proxy, str):
            self.proxy = ProxySettings(proxy)
        return self

    def set_wait_before_capture(self, milliseconds):
        # type: (int) -> Self
        self.wait_before_capture = milliseconds
        return self

    @deprecated.attribute("Configuration features are not supported")
    def set_features(self, *_):
        pass

    def set_user_test_id(self, user_test_id):
        # type: (Text) -> Self
        argument_guard.is_a(user_test_id, str)
        self.user_test_id = user_test_id
        return self

    def set_disable_nml_url_cache(self, disable_nml_url_cache):
        # type: (bool) -> Self
        argument_guard.is_a(disable_nml_url_cache, bool)
        self.disable_nml_url_cache = disable_nml_url_cache
        return self



from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING, List, Optional, Text, Tuple, TypeVar, Union, overload

import attr

from ..accessibility import AccessibilityRegionType
from ..geometry import AccessibilityRegion, Region
from ..layout_breakpoints_options import LayoutBreakpointsOptions
from ..ultrafastgrid import VisualGridOption
from ..utils import argument_guard
from ..validators import is_list_or_tuple, is_webelement
from .check_settings import CheckSettings, CheckSettingsValues
from .region import (
    AccessibilityRegionBySelector,
    DynamicRegionBySelector,
    FloatingRegionBySelector,
    RegionBySelector,
)
from .target_path import RegionLocator, TargetPath

if TYPE_CHECKING:
    from ..extract_text import OCRRegion
    from ..utils.custom_types import (
        FLOATING_VALUES,
        AnyWebElement,
        BySelector,
        CodedRegionPadding,
        CssSelector,
        FrameIndex,
        FrameNameOrId,
    )

BEFORE_CAPTURE_SCREENSHOT = "beforeCaptureScreenshot"


@attr.s
class FrameLocator(object):
    frame_locator = attr.ib(default=None)  # type: RegionLocator
    frame_name_or_id = attr.ib(default=None)  # type: FrameNameOrId
    frame_index = attr.ib(default=None)  # type: FrameIndex
    scroll_root_locator = attr.ib(default=None)  # type: RegionLocator


@attr.s
class LazyLoadOptions(object):
    scroll_length = attr.ib(default=None)  # type: Optional[int]
    waiting_time = attr.ib(default=None)  # type: Optional[int]
    max_amount_to_scroll = attr.ib(default=None)  # type: Optional[int]


@attr.s
class WebCheckSettingsValues(CheckSettingsValues):
    # hide_caret = attr.ib(init=False, default=None)
    scroll_root_locator = attr.ib(init=False, default=None)  # type: RegionLocator
    target_locator = attr.ib(init=False, default=None)  # type: RegionLocator
    frame_chain = attr.ib(init=False, factory=list)  # type: List[FrameLocator]

    script_hooks = attr.ib(factory=dict)  # type: dict
    visual_grid_options = attr.ib(default=())  # type: Tuple[VisualGridOption]
    disable_browser_fetching = attr.ib(default=None)  # type: Optional[bool]
    ocr_region = attr.ib(init=False, default=None)  # type: Optional[OCRRegion]
    layout_breakpoints = attr.ib(
        default=None
    )  # type: Union[bool, List[int], LayoutBreakpointsOptions, None]
    lazy_load = attr.ib(default=None)  # type: Optional[LazyLoadOptions]
    webview = attr.ib(default=None)  # type: Union[None, bool, Text]
    screenshot_mode = attr.ib(default=None)  # type: Optional[Text]

    @property
    def size_mode(self):
        if self.is_target_empty:
            if self.stitch_content:
                return "full-page"
            return "viewport"
        elif self.target_region:
            return "region"
        elif self.stitch_content:
            return "full-selector"
        return "selector"

    @property
    def is_target_empty(self):
        # type: () -> bool
        return self.target_region is None and self.target_locator is None


Self = TypeVar("Self", bound="WebCheckSettings")


@attr.s
class WebCheckSettings(CheckSettings):
    Values = WebCheckSettingsValues

    @overload  # noqa
    def layout(
        self,
        *by,  # type: BySelector
        **kwargs  # type: Union[Text, CodedRegionPadding]
    ):
        # type: (...) -> Self
        pass

    @overload  # noqa
    def layout(
        self,
        *element,  # type: AnyWebElement
        **kwargs  # type: Union[Text, CodedRegionPadding]
    ):
        # type: (...) -> Self
        pass

    @overload  # noqa
    def layout(
        self,
        *css_selector,  # type: CssSelector
        **kwargs  # type: Union[Text, CodedRegionPadding]
    ):
        # type: (...) -> Self
        pass

    @overload  # noqa
    def layout(
        self,
        *target_path,  # type: RegionLocator
        **kwargs  # type: Union[Text, CodedRegionPadding]
    ):
        # type: (...) -> Self
        pass

    @overload  # noqa
    def layout(
        self,
        *region,  # type: Region
        **kwargs  # type: Union[Text, CodedRegionPadding]
    ):
        # type: (...) -> Self
        pass

    def layout(self, *region, **kwargs):  # noqa
        """
        Adds one or more layout regions. And allow to set up paddings for them.

        :param kwargs: accepts `padding` parameter where there region paddings
         could be specified.
         Example of definition: `padding=dict(top=10, left=10, right=11, bottom=12)`
          where each keys are optional
        """
        return super(WebCheckSettings, self).layout(*region, **kwargs)

    @overload  # noqa
    def strict(
        self,
        *by,  # type: BySelector
        **kwargs  # type: Union[Text, CodedRegionPadding]
    ):
        # type: (...) -> Self
        pass

    @overload  # noqa
    def strict(
        self,
        *element,  # type: AnyWebElement
        **kwargs  # type: Union[Text, CodedRegionPadding]
    ):
        # type: (...) -> Self
        pass

    @overload  # noqa
    def strict(
        self,
        *css_selector,  # type: CssSelector
        **kwargs  # type: Union[Text, CodedRegionPadding]
    ):
        # type: (...) -> Self
        pass

    @overload  # noqa
    def strict(
        self,
        *target_path,  # type: RegionLocator
        **kwargs  # type: Union[Text, CodedRegionPadding]
    ):
        # type: (...) -> Self
        pass

    @overload  # noqa
    def strict(
        self,
        *region,  # type: Region
        **kwargs  # type: Union[Text, CodedRegionPadding]
    ):
        # type: (...) -> Self
        pass

    def strict(self, *region, **kwargs):  # noqa
        """
        Adds one or more strict regions. And allow to set up paddings for them.

        :param kwargs: accepts `padding` parameter where there region paddings
         could be specified.
         Example of definition: `padding=dict(top=10, left=10, right=11, bottom=12)`
          where each keys are optional
        """
        return super(WebCheckSettings, self).strict(*region, **kwargs)

    @overload  # noqa
    def content(
        self,
        *by,  # type: BySelector
        **kwargs  # type: Union[Text, CodedRegionPadding]
    ):
        # type: (...) -> Self
        pass

    @overload  # noqa
    def content(
        self,
        *element,  # type: AnyWebElement
        **kwargs  # type: Union[CodedRegionPadding, Text]
    ):
        # type: (...) -> Self
        pass

    @overload  # noqa
    def content(
        self,
        *css_selector,  # type: CssSelector
        **kwargs  # type: Union[CodedRegionPadding,Text]
    ):
        # type: (...) -> Self
        pass

    @overload  # noqa
    def content(
        self,
        *target_path,  # type: RegionLocator
        **kwargs  # type: Union[CodedRegionPadding, Text]
    ):
        # type: (...) -> Self
        pass

    @overload  # noqa
    def content(
        self,
        *region,  # type: Region
        **kwargs  # type: Union[CodedRegionPadding, Text]
    ):
        # type: (...) -> Self
        pass

    def content(self, *region, **kwargs):  # noqa
        """
        Adds one or more ignore colors regions. Optionally set up paddings for them.

        :param kwargs: accepts `padding` parameter that specifies padding for regions.
         Example of definition: `padding=dict(top=10, left=10, right=11, bottom=12)`
          where all keys are optional.

        This method is kept for backward compatibility.
        """
        return super(WebCheckSettings, self).ignore_colors(*region, **kwargs)

    @overload  # noqa
    def ignore_colors(
        self,
        *by,  # type: BySelector
        **kwargs  # type: Union[Text, CodedRegionPadding]
    ):
        # type: (...) -> Self
        pass

    @overload  # noqa
    def ignore_colors(
        self,
        *element,  # type: AnyWebElement
        **kwargs  # type: Union[CodedRegionPadding, Text]
    ):
        # type: (...) -> Self
        pass

    @overload  # noqa
    def ignore_colors(
        self,
        *css_selector,  # type: CssSelector
        **kwargs  # type: Union[CodedRegionPadding,Text]
    ):
        # type: (...) -> Self
        pass

    @overload  # noqa
    def ignore_colors(
        self,
        *target_path,  # type: RegionLocator
        **kwargs  # type: Union[CodedRegionPadding, Text]
    ):
        # type: (...) -> Self
        pass

    @overload  # noqa
    def ignore_colors(
        self,
        *region,  # type: Region
        **kwargs  # type: Union[CodedRegionPadding, Text]
    ):
        # type: (...) -> Self
        pass

    def ignore_colors(self, *region, **kwargs):  # noqa
        """
        Adds one or more ignore colors regions. Optionally set up paddings for them.

        :param kwargs: accepts `padding` parameter that specifies padding for regions.
         Example of definition: `padding=dict(top=10, left=10, right=11, bottom=12)`
          where all keys are optional
        """
        return super(WebCheckSettings, self).ignore_colors(*region, **kwargs)

    @overload  # noqa
    def ignore(
        self,
        *by,  # type: BySelector
        **kwargs  # type: Union[CodedRegionPadding, Text]
    ):
        # type: (...) -> Self
        pass

    @overload  # noqa
    def ignore(
        self,
        *element,  # type: AnyWebElement
        **kwargs  # type: Union[CodedRegionPadding, Text]
    ):
        # type: (...) -> Self
        pass

    @overload  # noqa
    def ignore(
        self,
        *css_selector,  # type: CssSelector
        **kwargs  # type: Union[CodedRegionPadding, Text]
    ):
        # type: (...) -> Self
        pass

    @overload  # noqa
    def ignore(
        self,
        *target_path,  # type: RegionLocator
        **kwargs  # type: Union[CodedRegionPadding, Text]
    ):
        # type: (...) -> Self
        pass

    @overload  # noqa
    def ignore(
        self,
        *region,  # type: Region
        **kwargs  # type: Union[CodedRegionPadding, Text]
    ):
        # type: (...) -> Self
        pass

    def ignore(self, *region, **kwargs):  # noqa
        """
        Adds one or more ignore regions. And allow to set up paddings for them.

        :param kwargs: accepts `padding` parameter where there region paddings
         could be specified.
         Example of definition: `padding=dict(top=10, left=10, right=11, bottom=12)`
          where each keys are optional
        """
        return super(WebCheckSettings, self).ignore(*region, **kwargs)

    @overload  # noqa
    def floating(self, max_offset, region):
        # type: (int, FLOATING_VALUES) -> Self
        pass

    @overload  # noqa
    def floating(
        self, region, max_up_offset, max_down_offset, max_left_offset, max_right_offset
    ):
        # type: (FLOATING_VALUES, int, int, int, int) -> Self
        pass

    def floating(self, *args):  # noqa
        return super(WebCheckSettings, self).floating(*args)

    @overload  # noqa
    def accessibility(self, region):  # noqa
        # type:(AccessibilityRegion) -> Self
        pass

    @overload  # noqa
    def accessibility(self, css_selector, type):  # noqa
        # type:(CssSelector, AccessibilityRegionType) -> Self
        pass

    @overload  # noqa
    def accessibility(self, target_path, type):  # noqa
        # type:(RegionLocator, AccessibilityRegionType) -> Self
        pass

    @overload  # noqa
    def accessibility(self, by, type):  # noqa
        # type:(BySelector, AccessibilityRegionType) -> Self
        pass

    @overload  # noqa
    def accessibility(self, element, type):  # noqa
        # type:(AnyWebElement, AccessibilityRegionType) -> Self
        pass

    @overload  # noqa
    def accessibility(self, region, type):  # noqa
        # type:(Region, AccessibilityRegionType) -> Self
        pass

    def accessibility(self, region, type=None):  # noqa
        # type:(...) -> Self
        return super(WebCheckSettings, self).accessibility(region, type)

    @overload  # noqa
    def region(self, region):  # noqa
        # type: (Region) -> Self
        pass

    @overload  # noqa
    def region(self, css_selector):  # noqa
        # type: (CssSelector) -> Self
        pass

    @overload  # noqa
    def region(self, target_path):  # noqa
        # type: (RegionLocator) -> Self
        pass

    @overload  # noqa
    def region(self, element):  # noqa
        # type: (AnyWebElement) -> Self
        pass

    @overload  # noqa
    def region(self, by):  # noqa
        # type: (BySelector) -> Self
        pass

    def region(self, region):  # noqa
        # type:(...) -> Self
        if isinstance(region, Region):
            self.values.target_region = region
        elif is_list_or_tuple(region):
            path = self.values.target_locator or TargetPath
            self.values.target_locator = path.region(*region)
        elif isinstance(region, str) or is_webelement(region):
            path = self.values.target_locator or TargetPath
            self.values.target_locator = path.region(region)
        elif isinstance(region, RegionLocator):
            self.values.target_locator = region
        else:
            raise TypeError("region method called with argument of unknown type!")
        return self

    @overload  # noqa
    def frame(self, frame_name_or_id):
        # type: (FrameNameOrId) -> Self
        pass

    @overload  # noqa
    def frame(self, element):
        # type: (AnyWebElement) -> Self
        pass

    @overload  # noqa
    def frame(self, index):
        # type: (FrameIndex) -> Self
        pass

    @overload  # noqa
    def frame(self, by):
        # type: (BySelector) -> Self
        pass

    @overload  # noqa
    def frame(self, target_path):
        # type: (RegionLocator) -> Self
        pass

    def frame(self, frame):  # noqa
        # type:(...) -> Self
        fl = FrameLocator()
        if isinstance(frame, int):
            fl.frame_index = frame
        elif isinstance(frame, str):
            fl.frame_name_or_id = frame
        elif is_webelement(frame):
            fl.frame_locator = TargetPath.frame(frame)
        elif is_list_or_tuple(frame):
            fl.frame_locator = TargetPath.frame(*frame)
        elif isinstance(frame, RegionLocator):
            fl.frame_locator = frame
        else:
            raise TypeError("frame method called with argument of unknown type!")
        self.values.frame_chain.append(fl)
        return self

    def before_render_screenshot_hook(self, hook):
        # type: (Text) -> Self
        self.values.script_hooks[BEFORE_CAPTURE_SCREENSHOT] = hook
        return self

    def _region_provider_from(self, region, method_name, padding, region_id):
        if isinstance(region, str) or is_webelement(region):
            return RegionBySelector(TargetPath.region(region), padding, region_id)
        elif is_list_or_tuple(region):
            return RegionBySelector(TargetPath.region(*region), padding, region_id)
        elif isinstance(region, RegionLocator):
            return RegionBySelector(region, padding, region_id)
        return super(WebCheckSettings, self)._region_provider_from(
            region, method_name, padding, region_id
        )

    def _set_scroll_root_locator(self, target_path):
        if len(self.values.frame_chain) == 0:
            self.values.scroll_root_locator = target_path
        else:
            self.values.frame_chain[-1].scroll_root_locator = target_path

    @overload  # noqa
    def scroll_root_element(self, element):
        # type: (AnyWebElement) -> Self
        pass

    @overload  # noqa
    def scroll_root_element(self, selector):
        # type: (CssSelector) -> Self
        pass

    @overload  # noqa
    def scroll_root_element(self, target_path):
        # type: (RegionLocator) -> Self
        pass

    @overload  # noqa
    def scroll_root_element(self, by):
        # type: (BySelector) -> Self
        pass

    def scroll_root_element(self, element_or_selector):  # noqa
        if isinstance(element_or_selector, str) or is_webelement(element_or_selector):
            self._set_scroll_root_locator(TargetPath.region(element_or_selector))
        elif is_list_or_tuple(element_or_selector):
            self._set_scroll_root_locator(TargetPath.region(*element_or_selector))
        elif isinstance(element_or_selector, RegionLocator):
            self._set_scroll_root_locator(element_or_selector)
        else:
            raise TypeError("Unsupported type of scroll root element")
        return self

    def _floating_provider_from(self, region, bounds):
        if isinstance(region, str) or is_webelement(region):
            return FloatingRegionBySelector(TargetPath.region(region), bounds)
        elif is_list_or_tuple(region):
            return FloatingRegionBySelector(TargetPath.region(*region), bounds)
        elif isinstance(region, RegionLocator):
            return FloatingRegionBySelector(region, bounds)
        return super(WebCheckSettings, self)._floating_provider_from(region, bounds)

    def _accessibility_provider_from(self, region, accessibility_region_type):
        if isinstance(region, str) or is_webelement(region):
            return AccessibilityRegionBySelector(
                TargetPath.region(region), accessibility_region_type
            )
        elif is_list_or_tuple(region):
            return AccessibilityRegionBySelector(
                TargetPath.region(*region), accessibility_region_type
            )
        elif isinstance(region, RegionLocator):
            return AccessibilityRegionBySelector(region, accessibility_region_type)
        return super(WebCheckSettings, self)._accessibility_provider_from(
            region, accessibility_region_type
        )

    def _dynamic_provider_from(self, region, dynamic_settings, padding=None):
        if isinstance(region, str) or is_webelement(region):
            return DynamicRegionBySelector(
                TargetPath.region(region), dynamic_settings, padding
            )
        elif is_list_or_tuple(region):
            return DynamicRegionBySelector(
                TargetPath.region(*region), dynamic_settings, padding
            )
        elif isinstance(region, RegionLocator):
            return DynamicRegionBySelector(region, dynamic_settings, padding)
        return super(WebCheckSettings, self)._dynamic_provider_from(
            region, dynamic_settings, padding
        )

    def visual_grid_options(self, *options):
        # type: (*VisualGridOption) -> Self
        argument_guard.are_(options, VisualGridOption)
        self.values.visual_grid_options = options
        return self

    def disable_browser_fetching(self, disable=True):
        # type: (bool) -> Self
        self.values.disable_browser_fetching = disable
        return self

    @overload
    def layout_breakpoints(
        self,
        enabled: bool,
        *,
        reload: Optional[bool] = None,
        height_breakpoints: Optional[bool] = None
    ) -> Self:
        pass

    @overload
    def layout_breakpoints(self, *breakpoints, **kwargs):
        # type: (*int, Optional[bool]) -> Self
        pass

    def layout_breakpoints(self, enabled_or_first, *rest, **kwargs):
        reload = kwargs.pop("reload", None)
        height_breakpoints = kwargs.pop("height_breakpoints", None)
        assert not kwargs, "Unexpected keyword arguments {}".format(kwargs)
        if isinstance(enabled_or_first, bool):
            assert (
                not rest
            ), "`reload` and `height_breakpoints` are keyword-only arguments"
            self.values.layout_breakpoints = LayoutBreakpointsOptions(
                enabled_or_first, reload, height_breakpoints
            )
        elif isinstance(enabled_or_first, int):
            self.values.layout_breakpoints = LayoutBreakpointsOptions(
                [enabled_or_first] + list(rest), reload, height_breakpoints
            )
        else:
            raise TypeError(
                "{} is not an instance of bool or int".format(enabled_or_first)
            )
        return self

    def lazy_load(
        self, scroll_length=None, waiting_time=None, max_amount_to_scroll=None
    ):
        # type: (Optional[int], Optional[int], Optional[int]) -> Self
        value = LazyLoadOptions(scroll_length, waiting_time, max_amount_to_scroll)
        self.values.lazy_load = value
        return self

    @overload
    def webview(self):
        # type: () -> Self
        pass

    @overload
    def webview(self, use_default):
        # type: (bool) -> Self
        pass

    @overload
    def webview(self, webview_id):
        # type: (Text) -> Self
        pass

    def webview(self, webview=True):
        self.values.webview = webview
        return self

    def use_system_screenshot(self, should_use=True):
        self.values.screenshot_mode = "default" if should_use else "applitools-lib"
        return self

    @property
    def is_check_window(self):
        # type: () -> bool
        return self.values.is_target_empty



from __future__ import absolute_import, division, print_function

from typing import TYPE_CHECKING, TypeVar, overload

from .target import Target
from .web_check_settings import WebCheckSettings

if TYPE_CHECKING:
    from ..utils.custom_types import (
        AnyWebElement,
        BySelector,
        CssSelector,
        FrameIndex,
        FrameNameOrId,
    )
    from . import Region
    from .target_path import RegionLocator

__all__ = ("WebTarget",)

Self = TypeVar("Self", bound="WebCheckSettings")


class WebTarget(Target):
    """
    Target for an eyes.check_window/region.
    """

    CheckSettings = WebCheckSettings

    @classmethod
    def window(cls):
        # type: () -> Self.CheckSettings
        return cls.CheckSettings()

    @classmethod
    @overload
    def region(cls, region):
        # type: (Region) -> Self.CheckSettings
        pass

    @classmethod
    @overload
    def region(cls, css_selector):
        # type: (CssSelector) -> Self.CheckSettings
        pass

    @classmethod
    @overload
    def region(cls, element):
        # type: (AnyWebElement) -> Self.CheckSettings
        pass

    @classmethod
    @overload
    def region(cls, by_selector):
        # type: (BySelector) -> Self.CheckSettings
        pass

    @classmethod
    @overload
    def region(cls, target_path):
        # type: (RegionLocator) -> Self.CheckSettings
        pass

    @classmethod
    def region(cls, region_or_image, rect=None):
        if rect:
            return Target.region(region_or_image, rect)
        else:
            return cls.CheckSettings().region(region_or_image)

    @classmethod
    @overload
    def frame(cls, frame_name_or_id):
        # type: (FrameNameOrId) -> Self.CheckSettings
        pass

    @classmethod
    @overload
    def frame(cls, frame_element):
        # type: (AnyWebElement) -> Self.CheckSettings
        pass

    @classmethod
    @overload
    def frame(cls, frame_index):
        # type: (FrameIndex) -> Self.CheckSettings
        pass

    @classmethod
    @overload
    def frame(cls, frame_by_selector):
        # type: (BySelector) -> Self.CheckSettings
        pass

    @classmethod
    @overload
    def frame(cls, target_path):
        # type: (RegionLocator) -> Self.CheckSettings
        pass

    @classmethod
    def frame(cls, frame):
        return cls.CheckSettings().frame(frame)



from __future__ import absolute_import, division, print_function

from typing import Any, Dict, List, Optional, Text

import attr

from ..geometry import RectangleSize
from ..selenium.misc import BrowserType
from .config import (
    AndroidVersion,
    DeviceName,
    IosDeviceName,
    IosVersion,
    ScreenOrientation,
)


class IRenderBrowserInfo(object):
    viewport_size = None
    width = 0
    height = 0


@attr.s
class EmulationBaseInfo(object):
    device_name = attr.ib()  # type: DeviceName
    screen_orientation = attr.ib()  # type: ScreenOrientation


@attr.s(hash=True)
class ChromeEmulationInfo(IRenderBrowserInfo):
    device_name = attr.ib(converter=DeviceName)
    screen_orientation = attr.ib(
        default=ScreenOrientation.PORTRAIT, converter=ScreenOrientation
    )  # type: ScreenOrientation
    # backward compatibility with sdk versions<5
    baseline_env_name = attr.ib(default=None)
    browser = BrowserType.CHROME.value
    platform = "linux"


def _convert_version(cls):
    def inner(value):
        if value is None:
            return None
        elif isinstance(value, str):
            try:
                return cls(value)
            except ValueError:
                return value
        elif isinstance(value, cls):
            return value
        else:
            raise ValueError("Invalid {} version type: {}", cls.__name__, type(value))

    return inner


@attr.s(hash=True)
class IosDeviceInfo(IRenderBrowserInfo):
    device_name = attr.ib(
        converter=lambda obj: getattr(obj, "value", obj)
    )  # type: Text | IosDeviceName
    screen_orientation = attr.ib(
        default=ScreenOrientation.PORTRAIT, converter=ScreenOrientation
    )  # type: Optional[ScreenOrientation]
    ios_version = attr.ib(
        default=None, converter=attr.converters.optional(_convert_version(IosVersion))
    )  # type: Optional[IosVersion | Text]
    # backward compatibility with sdk versions<5
    baseline_env_name = attr.ib(default=None)
    browser = BrowserType.SAFARI.value
    platform = "ios"


@attr.s(hash=True)
class AndroidDeviceInfo(IRenderBrowserInfo):
    device_name = attr.ib(
        converter=lambda obj: getattr(obj, "value", obj)
    )  # type: Text | AndroidDeviceName
    screen_orientation = attr.ib(
        default=ScreenOrientation.PORTRAIT, converter=ScreenOrientation
    )  # type: Optional[ScreenOrientation]
    android_version = attr.ib(
        default=None,
        converter=attr.converters.optional(_convert_version(AndroidVersion)),
    )  # type: Optional[AndroidVersion | Text]
    # backward compatibility with sdk versions<5
    baseline_env_name = attr.ib(default=None)
    browser = None
    platform = "android"


@attr.s(hash=True)
class DesktopBrowserInfo(IRenderBrowserInfo):
    width = attr.ib(type=int)
    height = attr.ib(type=int)
    browser_type = attr.ib(default=BrowserType.CHROME, converter=BrowserType)
    # backward compatibility with sdk versions<5
    baseline_env_name = attr.ib(default=None)

    @property
    def browser(self):
        # type: () -> Text
        return self.browser_type.value

    @property
    def platform(self):
        # type: () -> Text
        if self.browser_type in [
            BrowserType.IE_10,
            BrowserType.IE_11,
            BrowserType.EDGE_LEGACY,
        ]:
            return "windows"
        elif self.browser_type in [
            BrowserType.SAFARI,
            BrowserType.SAFARI_ONE_VERSION_BACK,
            BrowserType.SAFARI_TWO_VERSIONS_BACK,
        ]:
            return "mac os x"
        return "linux"


@attr.s(hash=True)
class EnvironmentInfo(IRenderBrowserInfo):
    render_environment_id = attr.ib(default=None)  # type: Optional[Text]
    ec_session_id = attr.ib(default=None)  # type: Optional[Text]
    os = attr.ib(default=None)  # type: Optional[Text]
    os_info = attr.ib(default=None)  # type: Optional[Text]
    hosting_app = attr.ib(default=None)  # type: Optional[Text]
    hosting_app_info = attr.ib(default=None)  # type: Optional[Text]
    device_name = attr.ib(default=None)  # type: Optional[Text]
    viewport_size = attr.ib(default=None)  # type: Optional[RectangleSize]
    user_agent = attr.ib(default=None)  # type: Optional[Text]
    renderer = attr.ib(default=None)  # type: Optional[Dict[Text, Any]]
    raw_environment = attr.ib(default=None)  # type: Optional[Dict[Text, Any]]
    properties = attr.ib(default=None)  # type: Optional[List[Dict[Text, Any]]]



from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING, List, Optional, Text, Tuple, Union, overload

from .. import common
from ..common.command_executor import CommandExecutor
from ..common.schema import demarshal_locate_result, demarshal_test_results
from . import EyesError, RectangleSize, deprecated
from .config import Configuration
from .extract_text import ImageOCRRegion
from .fluent.check_settings import CheckSettings
from .fluent.images_check_settings import ImagesCheckSettings
from .fluent.target import Target
from .fluent.web_target import WebTarget
from .runner import EyesRunner, log_session_results_and_raise_exception
from .schema import demarshal_locate_text_result
from .target import ImageTarget

if TYPE_CHECKING:
    from typing import ByteString, TypeVar

    from . import Region, TestResults
    from .extract_text import PATTERN_TEXT_REGIONS, Image, OCRRegion, TextRegionSettings
    from .locators import LOCATORS_TYPE, VisualLocatorSettings
    from .optional_deps import (
        ElementHandle,
        Page,
        PlaywrightLocator,
        WebDriver,
        WebElement,
    )
    from .utils.custom_types import FrameReference, ViewPort

    Self = TypeVar("Self", bound="EyesBase")  # typedef


class EyesBase(object):
    _Configuration = Configuration
    _DefaultRunner = EyesRunner

    def __init__(self, runner):
        # type: (Optional[Self._DefaultRunner]) -> None
        self._eyes_ref = None
        # get_results should be possible to call after close so we can't use
        # only _eyes_ref to determine openness anymore.
        self._is_open = False
        self._object_registry = None
        self.configuration = self._Configuration()
        self._runner = runner or self._DefaultRunner()
        self._commands = self._runner._commands  # noqa

    @property
    def configure(self):
        # type: () -> Self._Configuration
        return self.configuration

    @configure.setter
    def configure(self, value):
        self.configuration = value

    def get_configuration(self):
        # type:() -> Self._Configuration
        return self.configuration.clone()

    def set_configuration(self, configuration):
        # type:(Self._Configuration) -> None
        self.configuration = configuration.clone()

    @property
    def is_open(self):
        # type: () -> bool
        return self._is_open

    @property
    def base_agent_id(self):
        # type: () -> Text
        """
        Must return version of SDK. (e.g. selenium, visualgrid) in next format:
            "eyes.{package}.python/{lib_version}"
        """
        return "{}/{}".format(self._runner.BASE_AGENT_ID, common.__version__)

    @property
    def full_agent_id(self):
        # type: () -> Text
        """
        Gets the agent id, which identifies the current library using the SDK.

        """
        if self.configuration.agent_id is None:
            return self.base_agent_id
        else:
            return "{} [{}]".format(self.configuration.agent_id, self.base_agent_id)

    @overload
    def check(self, name, check_settings):
        # type: (Text, ImagesCheckSettings) -> bool
        pass

    @overload
    def check(self, check_settings):
        # type: (ImagesCheckSettings) -> bool
        pass

    def check(self, check_settings, name=None):
        # type: (ImagesCheckSettings, Optional[Text]) -> None
        if isinstance(name, ImagesCheckSettings) or isinstance(check_settings, str):
            check_settings, name = name, check_settings
        if name:
            check_settings = check_settings.with_name(name)

        self._commands.eyes_check(
            self._object_registry,
            self._eyes_ref,
            ImageTarget(check_settings.values.image, check_settings.values.dom),
            check_settings,
            self.configuration,
        )

    def check_image(self, image, tag=None):
        # type: (Union[ByteString, Text, Image], Optional[Text]) -> bool
        return self.check(tag, Target.image(image))

    def locate_text(self, config):
        # type: (TextRegionSettings) -> PATTERN_TEXT_REGIONS
        return self.extract_text_regions(config)

    def extract_text_regions(self, config):
        # type: (TextRegionSettings) -> PATTERN_TEXT_REGIONS
        image = config._image  # noqa
        assert image, "TextRegionSettings has to specify the image to extract text from"
        result = self._commands.core_locate_text(
            ImageTarget(config._image),  # noqa
            config,
            self.configuration,
        )
        return demarshal_locate_text_result(result)

    def extract_text(self, *regions):
        # type: (*ImageOCRRegion) -> List[Text]
        image = regions[0].image
        assert all(r.image == image for r in regions), "All images same"
        return self._commands.core_extract_text(
            ImageTarget(image),
            regions,
            self.configuration,
        )

    def locate(self, visual_locator_settings):
        # type: (VisualLocatorSettings) -> LOCATORS_TYPE
        results = self._commands.core_locate(
            ImageTarget(visual_locator_settings.values.image),
            visual_locator_settings,
            self.configuration,
        )
        return demarshal_locate_result(results)

    def get_results(self, raise_ex=True):
        # type: (bool) -> List[TestResults]
        """
        Wait and return results of checks made after eyes were opened.

        :param raise_ex: raise an exception if at least one check has failed.
        """
        if not self._eyes_ref:
            raise EyesError("Eyes was never open")
        results = self._commands.eyes_get_results(self._eyes_ref, raise_ex)
        if results is not None:
            results = demarshal_test_results(results, self.configuration)
            if results:  # eyes are already aborted by closed runner
                for r in results:
                    log_session_results_and_raise_exception(raise_ex, r)
                return results
        return []

    def close(self, raise_ex=True):
        # type: (bool) -> Optional[TestResults]
        """
        Ends the test.

        :param raise_ex: If true, an exception will be raised for failed/new tests.
        :return: The test results.
        """
        return self._close(raise_ex, True)

    def abort(self):
        # type: () -> Optional[TestResults]
        return self._abort(True)

    def _open(
        self,
        app_name,  # type: Optional[Text]
        test_name,  # type: Optional[Text]
        viewport_size,  # type: Optional[ViewPort]
        target,  # type: Union[None, WebDriver, Page]
    ):
        # type: (...) -> None
        if not self.configuration.is_disabled:
            if app_name is not None:
                self.configuration.app_name = app_name
            if test_name is not None:
                self.configuration.test_name = test_name
            if viewport_size is not None:
                self.configuration.viewport_size = viewport_size
            if self.configuration.app_name is None:
                raise ValueError(
                    "app_name should be set via configuration or an argument"
                )
            if self.configuration.test_name is None:
                raise ValueError(
                    "test_name should be set via configuration or an argument"
                )

            self._runner._set_connection_config(self.configuration)  # noqa, friend
            self._object_registry = self._runner.Protocol.object_registry()
            self._eyes_ref = self._commands.manager_open_eyes(
                self._object_registry,
                self._runner._ref,  # noqa
                target,
                self.configuration,
            )
            self._is_open = True

    def _close(self, raise_ex, wait_result):
        # type: (bool, bool) -> Optional[TestResults]
        if self.configuration.is_disabled:
            return None
        if not self.is_open:
            raise EyesError("Eyes not open")
        self._commands.eyes_close(
            self._object_registry, self._eyes_ref, raise_ex, self.configuration
        )
        results = self.get_results(raise_ex) if wait_result else []
        self._object_registry = None
        self._is_open = False
        # Original interface returns just the first result
        return results[0] if results else None

    def _abort(self, wait_result):
        # type: (bool) -> Optional[TestResults]
        if self.configuration.is_disabled:
            return None
        elif self.is_open:
            self._commands.eyes_abort(self._object_registry, self._eyes_ref)
            results = self.get_results(False) if wait_result else None
            self._object_registry = None
            self._is_open = False
            # Original interface returns just the first result
            return results[0] if results else None

    def __getattr__(self, item):
        return getattr(self.configuration, item)

    def __setattr__(self, key, value):
        if "configuration" in vars(self) and (
            key in vars(self.configuration)
            or key in ("match_level", "ignore_displacements")
        ):
            return setattr(self.configuration, key, value)
        else:
            return super(EyesBase, self).__setattr__(key, value)


class WebEyes(EyesBase):
    Target = WebTarget

    def __init__(self, runner=None):
        # type: (Union[None, EyesRunner, Text]) -> None
        if isinstance(runner, str):
            server_url = runner
            runner = None
        else:
            server_url = None
        super(WebEyes, self).__init__(runner)
        self.driver = None
        if server_url:
            self.configuration.set_server_url(server_url)

    def open(
        self,
        driver,  # type: Union[WebDriver, Page]
        app_name=None,  # type: Optional[Text]
        test_name=None,  # type: Optional[Text]
        viewport_size=None,  # type: Optional[ViewPort]
    ):
        # type: (...) -> Union[WebDriver, Page]
        self._open(app_name, test_name, viewport_size, driver)
        self.driver = driver
        return driver

    @overload
    def check(self, name, check_settings):
        # type: (Text, CheckSettings) -> None
        """
        Takes a snapshot and matches it with the expected output.

        :param name: The name of the tag.
        :param check_settings: target which area of the window to check.
        """
        pass

    @overload
    def check(self, check_settings):
        # type: (CheckSettings) -> None
        """
        Takes a snapshot and matches it with the expected output.

        :param check_settings: target which area of the window to check.
        """
        pass

    def check(self, check_settings, name=None):
        # type: (CheckSettings, Optional[Text]) -> None
        if isinstance(name, CheckSettings) or isinstance(check_settings, str):
            check_settings, name = name, check_settings
        if type(check_settings) is ImagesCheckSettings:
            return super(WebEyes, self).check(check_settings, name)
        if check_settings is None:
            check_settings = self.Target.window()
        if name:
            check_settings = check_settings.with_name(name)

        if self.configuration.is_disabled:
            return
        if not self.is_open:
            self.abort()
            raise EyesError("you must call open() before checking")

        self._commands.eyes_check(
            self._object_registry,
            self._eyes_ref,
            None,
            check_settings,
            self.configuration,
        )

    def locate(self, visual_locator_settings):
        # type: (VisualLocatorSettings) -> LOCATORS_TYPE
        if visual_locator_settings.values.image:  # noqa
            return super(WebEyes, self).locate(visual_locator_settings)
        else:
            results = self._commands.core_locate(
                self.driver,
                visual_locator_settings,
                self.configuration,
            )
            return demarshal_locate_result(results)

    def extract_text(self, *regions):
        # type: (*OCRRegion) -> List[Text]
        first_region = regions[0]
        assert all(
            type(r) is type(first_region) for r in regions
        ), "Regions should be same type"
        if isinstance(first_region, ImageOCRRegion):
            return super(WebEyes, self).extract_text(*regions)
        else:
            return self._commands.core_extract_text(
                self.driver,
                regions,
                self.configuration,
            )

    def extract_text_regions(self, config):
        # type: (TextRegionSettings) -> PATTERN_TEXT_REGIONS
        if config._image:  # noqa
            return super(WebEyes, self).extract_text_regions(config)
        else:
            return self._commands.core_locate_text(
                self.driver, config, self.configuration
            )

    def close(self, raise_ex=True):
        # type: (bool) -> Optional[TestResults]
        """
        Ends the test.

        :param raise_ex: If true, an exception will be raised for failed/new tests.
        :return: The test results.
        """
        self.driver = None
        return super(WebEyes, self).close(raise_ex)

    def close_async(self):
        # type: () -> Optional[TestResults]
        self.driver = None
        return self._close(False, False)

    def abort(self):
        # type: () -> Optional[TestResults]
        """
        If a test is running, aborts it. Otherwise, does nothing.
        """
        self.driver = None
        return super(WebEyes, self).abort()

    def abort_async(self):
        self.driver = None
        return self._abort(False)

    @deprecated.attribute("use `abort()` instead")
    def abort_if_not_closed(self):
        return self.abort()

    @classmethod
    def get_viewport_size(cls, driver):
        # type: (Union[WebDriver, Page]) -> RectangleSize
        cmd = CommandExecutor.get_instance(cls._DefaultRunner.Protocol)
        result = cmd.core_get_viewport_size(driver)
        return RectangleSize.from_(result)

    @classmethod
    def set_viewport_size(cls, driver, viewport_size):
        # type: (Union[WebDriver, Page], ViewPort) -> None
        cmd = CommandExecutor.get_instance(cls._DefaultRunner.Protocol)
        cmd.core_set_viewport_size(driver, viewport_size)

    def check_window(self, tag=None, match_timeout=-1, fully=None):
        # type: (Optional[Text], int, Optional[bool]) -> None
        """
        Takes a snapshot of the application under test and matches it with the expected
         output.

        :param tag: An optional tag to be associated with the snapshot.
        :param match_timeout:  The amount of time to retry matching (milliseconds)
        :param fully: Defines that the screenshot will contain the entire window.
        """
        return self.check(tag, self.Target.window().timeout(match_timeout).fully(fully))

    def check_frame(self, frame_reference, tag=None, match_timeout=-1):
        # type: (FrameReference, Optional[Text], int) -> None
        """
        Check frame.

        :param frame_reference: The name or id of the frame to check. (The same
                name/id as would be used in a call to driver.switch_to.frame()).
        :param tag: An optional tag to be associated with the match.
        :param match_timeout: The amount of time to retry matching. (Milliseconds)
        """
        return self.check(
            tag, self.Target.frame(frame_reference).fully().timeout(match_timeout)
        )

    def check_region(
        self,
        region,  # type: Union[Region, Text, List, Tuple, WebElement]
        tag=None,  # type: Optional[Text]
        match_timeout=-1,  # type: int
        stitch_content=False,  # type: bool
    ):
        # type: (...) -> None
        """
        Takes a snapshot of the given region from the browser using the web driver
        and matches it with the expected output. If the current context is a frame,
        the region is offsetted relative to the frame.

        :param region: The region which will be visually validated. The coordinates are
                       relative to the viewport of the current frame.
        :param tag: Description of the visual validation checkpoint.
        :param match_timeout: Timeout for the visual validation checkpoint
                              (milliseconds).
        :param stitch_content: If `True`, stitch the internal content of the region
        """
        return self.check(
            tag,
            self.Target.region(region).timeout(match_timeout).fully(stitch_content),
        )

    def check_element(
        self,
        element,
        # type: Union[Text,List,Tuple,WebElement,PlaywrightLocator,ElementHandle]
        tag=None,  # type: Optional[Text]
        match_timeout=-1,  # type: int
    ):
        # type: (...) -> None
        """
        Takes a snapshot of the given region from the browser using the web driver
        and matches it with the expected output. If the current context is a frame,
        the region is offsetted relative to the frame.

        :param element: The element to check.
        :param tag: Description of the visual validation checkpoint.
        :param match_timeout: Timeout for the visual validation checkpoint
                              (milliseconds).
        """
        return self.check(
            tag,
            self.Target.region(element).timeout(match_timeout).fully(),
        )

    def check_region_in_frame(
        self,
        frame_reference,  # type: FrameReference
        region,
        # type: Union[Region,Text,List,Tuple,WebElement,PlaywrightLocator,ElementHandle]
        tag=None,  # type: Optional[Text]
        match_timeout=-1,  # type: int
        stitch_content=False,  # type: bool
    ):
        # type: (...) -> None
        """
        Checks a region within a frame, and returns to the current frame.

        :param frame_reference: A reference to the frame in which the region
                                should be checked.
        :param region: Specifying the region to check inside the frame.
        :param tag: Description of the visual validation checkpoint.
        :param match_timeout: Timeout for the visual validation checkpoint
                              (milliseconds).
        :param stitch_content: If `True`, stitch the internal content of the region
        """
        return self.check(
            tag,
            self.Target.region(region)
            .frame(frame_reference)
            .stitch_content(stitch_content)
            .timeout(match_timeout),
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.driver = None
        if exc_type:
            self._abort(self._runner.AUTO_CLOSE_MODE_SYNC)
        else:
            self._close(True, self._runner.AUTO_CLOSE_MODE_SYNC)



from __future__ import absolute_import, division, print_function

from typing import TYPE_CHECKING

import attr

if TYPE_CHECKING:
    from typing import Optional, Text


@attr.s
class ImageTarget(object):
    image = attr.ib(default=None)  # type: Optional[Text]
    dom = attr.ib(default=None)  # type: Optional[Text]



from __future__ import absolute_import, division, print_function, unicode_literals

from applitools.common import deprecated
from applitools.common.fluent.check_settings import CheckSettings, CheckSettingsValues

deprecated.module(__name__)
__all__ = "CheckSettings", "CheckSettingsValues"



from __future__ import absolute_import, division, print_function, unicode_literals

from ..common import deprecated
from ..common.cut import FixedCutProvider, NullCutProvider, UnscaledFixedCutProvider

deprecated.module(__name__)
__all__ = "FixedCutProvider", "NullCutProvider", "UnscaledFixedCutProvider"



from __future__ import absolute_import, division, print_function

import enum
from typing import TYPE_CHECKING

from . import (
    ChromeEmulationInfo,
    DesktopBrowserInfo,
    DiffsFoundError,
    IosDeviceInfo,
    NewTestError,
    TestFailedError,
)
from .context_vars import get_object_registry
from .errors import USDKFailure
from .extract_text import OCRRegion
from .fluent.region import (
    AccessibilityRegionByRectangle,
    AccessibilityRegionBySelector,
    DynamicRegionByRectangle,
    DynamicRegionBySelector,
    FloatingRegionByRectangle,
    FloatingRegionBySelector,
    RegionByRectangle,
    RegionBySelector,
)
from .fluent.target_path import RegionLocator
from .layout_breakpoints_options import LayoutBreakpointsOptions as LBO
from .mmallow import Dict, Raw
from .optional_deps import StaleElementReferenceException
from .ultrafastgrid import AndroidDeviceInfo

if TYPE_CHECKING:
    import typing as t

    from . import config as cfg
    from . import ultrafastgrid as ufg
    from .fluent import region, target_path
    from .fluent import web_check_settings as cs
    from .selenium import misc as selenium_misc


class Enum(Raw):
    # Customize the error message so that it matches the expected output.
    default_error_messages = {"invalid": "is not a valid {enum_name}"}

    def __init__(self, enum_type, data_key=None, attribute=None):
        # enum_type is the expected enum (e.g. StitchMode)
        super().__init__(data_key=data_key, attribute=attribute)
        self.enum_type = enum_type

    def _serialize(self, value, attr, obj, *args, **kwargs):
        """
        Converts the enum instance (or a raw compatible value) into its underlying value.
        Accepts:
          - None: returns None.
          - An instance of the expected enum type: returns its .value.
          - An instance of a different enum: attempts to convert it using its .value.
          - A raw value: attempts to convert it to the expected enum.
        If conversion fails, self.fail() is called.
        """
        if value is None:
            return None
        try:
            if isinstance(value, self.enum_type):
                # Value is already an instance of the expected enum.
                return value.value
            elif isinstance(value, enum.Enum):
                # Value is an enum of a different type.
                # * robotframework library defines customized enums like RobotStitchMode
                # Try to create an instance of the expected enum using its .value.
                return self.enum_type(value.value).value
            else:
                # Accept raw values.
                return self.enum_type(value).value
        except ValueError:
            raise self.make_error("invalid", enum_name=self.enum_type.__name__)

    def _deserialize(self, value, attr, data, *args, **kwargs):
        if value is None:
            return None
        try:
            return self.enum_type(value)
        except ValueError:
            raise self.make_error("invalid", enum_name=self.enum_type.__name__)


class Error(Raw):
    def _deserialize(self, value, attr, data, *args, **kwargs):
        # type: (dict, t.Any, t.Any, *t.Any) -> Exception
        return demarshal_error(value)


class DebugScreenshots(Raw):
    _CHECK_ATTRIBUTE = False

    def _serialize(self, _, __, config, *args, **kwargs):
        # type: (t.Any, t.Any, cfg.Configuration, *t.Any, **t.Any) -> dict
        from .schema import DebugScreenshotHandler

        if config.save_debug_screenshots:
            return DebugScreenshotHandler().dump(config)


class EnvironmentRaw(Raw):
    _CHECK_ATTRIBUTE = False

    def _serialize(self, _, __, config, *args, **kwargs):
        # type: (t.Any, t.Any, cfg.Configuration, *t.Any, **t.Any) -> dict
        from .schema import Environment

        return Environment().dump(config)


class VisualGridOptions(Raw):
    def _serialize(self, value, *args, **kwargs):
        # type: (t.Optional[t.List[ufg.VisualGridOption]], *t.Any, **t.Any) -> t.Optional[dict]
        if value is not None:
            return {r.key: r.value for r in value}
        else:
            return None


class ElementReference(Dict):
    def _serialize(self, locator, *args, **kwargs):
        # type: (t.Any, target_path.TargetPathLocator, *t.Any, **t.Any) -> t.Optional[dict]
        return None if locator is None else locator.to_dict(get_object_registry())


class MobileOptions(Dict):
    def _serialize(self, obj, *args, **kwargs):
        # type: (t.Any, selenium_misc.MobileOptions, t.Any) -> t.Optional[dict]
        if getattr(obj, "keep_navigation_bar", None) is not None:
            return {"keepNavigationBar": obj.keep_navigation_bar}
        return None


class FrameReference(Raw):
    _CHECK_ATTRIBUTE = False

    def _serialize(self, _, __, frame, *args, **kwargs):
        # type: (t.Any, t.Any, cs.FrameLocator,*t.Any, **t.Any) -> t.Union[int, t.Text, dict]
        if frame.frame_index is not None:
            return frame.frame_index
        elif frame.frame_name_or_id is not None:
            return frame.frame_name_or_id
        else:
            return frame.frame_locator.to_dict(get_object_registry())


class LayoutBreakpoints(Raw):
    """This custom field serializer is needed to provide backward compatibility with
    code that explicitly sets value of layout_breakpoints configuration attribute"""

    def _serialize(self, lbo, *args, **kwargs):
        # type: (t.Union[bool, list, tuple, LBO], *t.Any, **t.Any) -> dict
        if isinstance(lbo, (bool, list)):
            lbo = LBO(lbo)
        elif isinstance(lbo, tuple):
            lbo = LBO(list(lbo))
        from .schema import LayoutBreakpointsOptions

        res = LayoutBreakpointsOptions().dump(lbo)
        return res


class NormalizationRaw(Raw):
    _CHECK_ATTRIBUTE = False

    def _serialize(self, _, __, config, *args, **kwargs):
        from .schema import Normalization

        return Normalization().dump(config)


class StitchOverlap(Raw):
    def _serialize(self, value, *args, **kwargs):
        # type: (t.Any, int, *t.Any, **t.Any) -> dict
        if value is not None:
            return {"bottom": value}


class TargetReference(Raw):
    _CHECK_ATTRIBUTE = False  # it might be target_locator or target_region

    def _serialize(self, _, __, check_settings, *args, **kwargs):
        # type: (t.Any, t.Any, cs.WebCheckSettingsValues, *t.Any, **t.Any) -> t.Optional[dict]
        if check_settings.target_locator:
            return check_settings.target_locator.to_dict(get_object_registry())
        elif check_settings.target_region:
            from .schema import Region

            return Region().dump(check_settings.target_region)
        else:
            return None


class RegionReference(Raw):
    _CHECK_ATTRIBUTE = False

    def _serialize(self, _, __, obj, *args, **kwargs):
        # type: (t.Any, t.Any, t.Union[region.GetRegion, OCRRegion], *t.Any, **t.Any) -> dict
        from .schema import Region

        if isinstance(
            obj,
            (
                RegionBySelector,
                FloatingRegionBySelector,
                AccessibilityRegionBySelector,
                DynamicRegionBySelector,
            ),
        ):
            return obj._target_path.to_dict(get_object_registry())  # noqa
        elif isinstance(obj, RegionByRectangle):
            return Region().dump(obj._region)  # noqa
        elif isinstance(
            obj,
            (
                FloatingRegionByRectangle,
                AccessibilityRegionByRectangle,
                DynamicRegionByRectangle,
            ),
        ):
            return Region().dump(obj._rect)  # noqa
        elif isinstance(obj, OCRRegion):
            if isinstance(obj.target, RegionLocator):
                return obj.target.to_dict(get_object_registry())
            else:
                return Region().dump(obj.target)
        else:
            raise RuntimeError("Unexpected region type", type(obj))


class BrowserInfo(Raw):
    def _serialize(self, value, *args, **kwargs):
        # type: (ufg.IRenderBrowserInfo, *t.Any, **t.Any) -> dict
        if isinstance(value, DesktopBrowserInfo):
            from .schema import DesktopBrowserRenderer

            return DesktopBrowserRenderer().dump(value)
        elif isinstance(value, ChromeEmulationInfo):
            from .schema import ChromeEmulationRenderer

            return {"chromeEmulationInfo": ChromeEmulationRenderer().dump(value)}
        elif isinstance(value, IosDeviceInfo):
            from .schema import IosDeviceRenderer

            return {"iosDeviceInfo": IosDeviceRenderer().dump(value)}
        elif isinstance(value, AndroidDeviceInfo):
            from .schema import AndroidDeviceRenderer

            return {"androidDeviceInfo": AndroidDeviceRenderer().dump(value)}
        else:
            raise RuntimeError("Unexpected BrowserInfo type", type(value))

    def _deserialize(self, value, *args, **kwargs):
        # type: (t.Optional[dict], *t.Any, **t.Any) -> t.Optional[ufg.IRenderBrowserInfo]
        if value is None or "requested" not in value:
            return None
        value = value["requested"]
        if "iosDeviceInfo" in value:
            from .schema import IosDeviceRenderer

            return IosDeviceRenderer().load(value["iosDeviceInfo"])
        elif "chromeEmulationInfo" in value:
            from .schema import ChromeEmulationRenderer

            return ChromeEmulationRenderer().load(value["chromeEmulationInfo"])
        elif "androidDeviceInfo" in value:
            from .schema import AndroidDeviceRenderer

            return AndroidDeviceRenderer().load(value["androidDeviceInfo"])
        elif "environment" in value:
            from .schema import EnvironmentRenderer

            return EnvironmentRenderer().load(value["environment"])
        elif "name" in value and "width" in value and "height" in value:
            from .schema import DesktopBrowserRenderer

            return DesktopBrowserRenderer().load(value)
        else:
            # According to core team unknown types should be ignored
            return None


def demarshal_error(error_dict):
    # type: (dict) -> Exception
    message = error_dict["message"]
    if message.startswith("stale element reference"):
        return StaleElementReferenceException(message)
    elif error_dict.get("reason") in _matching_failures:
        return _matching_failures[error_dict["reason"]](message)
    else:
        stack = error_dict["stack"]
        if message:  # Sometimes when internal error occurs the message is empty
            # There is usually a copy of message in stack trace too, remove it
            stack = stack.split(message)[-1].strip("\n")
        return USDKFailure(message, stack)


_matching_failures = {
    "test different": DiffsFoundError,
    "test failed": TestFailedError,
    "test new": NewTestError,
}



from __future__ import absolute_import, division, print_function

from copy import deepcopy
from typing import TYPE_CHECKING, List, Optional, Text, Tuple, Union, overload

import attr

from ..config import Configuration as ConfigurationBase
from ..layout_breakpoints_options import LayoutBreakpointsOptions
from ..ultrafastgrid import (
    AndroidDeviceInfo,
    AndroidDeviceTarget,
    AndroidMultiDeviceTarget,
    ChromeEmulationInfo,
    DesktopBrowserInfo,
    DeviceTarget,
    IosDeviceInfo,
    IosDeviceTarget,
    IosMultiDeviceTarget,
    IRenderBrowserInfo,
    ScreenOrientation,
    VisualGridOption,
)
from ..utils import argument_guard
from ..validators import is_list_or_tuple
from .misc import BrowserType, MobileOptions, StitchMode

if TYPE_CHECKING:
    from ..cut import CutProvider
    from ..ultrafastgrid import DeviceName

__all__ = ("Configuration",)


@attr.s
class Configuration(ConfigurationBase):
    force_full_page_screenshot = attr.ib(default=None)  # type: bool
    wait_before_screenshots = attr.ib(
        default=None,
    )  # type: Optional[int]  # ms
    stitch_mode = attr.ib(default=None)  # type: Optional[StitchMode]
    hide_scrollbars = attr.ib(default=None)  # type: bool
    hide_caret = attr.ib(default=None)  # type: Optional[bool]
    # Indicates that a mobile simulator is being used
    is_simulator = attr.ib(default=None)  # type: Optional[bool]

    # Rendering Configuration
    browsers_info = attr.ib(init=False, factory=list)  # type: List[IRenderBrowserInfo]
    visual_grid_options = attr.ib(
        default=None
    )  # type: Optional[Tuple[VisualGridOption]]
    disable_browser_fetching = attr.ib(default=None)  # type: Optional[bool]
    enable_cross_origin_rendering = attr.ib(default=None)  # type: Optional[bool]
    dont_use_cookies = attr.ib(default=None)  # type: Optional[bool]
    dont_close_batches = attr.ib(default=None)  # type: Optional[bool]
    layout_breakpoints = attr.ib(
        default=None
    )  # type: Union[bool, List[int], LayoutBreakpointsOptions, None]
    scale_ratio = attr.ib(default=None)  # type: Optional[float]
    cut_provider = attr.ib(default=None)  # type: Optional[CutProvider]
    rotation = attr.ib(default=None)  # type: Optional[int]
    mobile_options = attr.ib(default=None)  # type: Optional[MobileOptions]

    def set_force_full_page_screenshot(self, force_full_page_screenshot):
        # type: (bool) -> Configuration
        self.force_full_page_screenshot = force_full_page_screenshot
        return self

    def set_wait_before_screenshots(self, wait_before_screenshots):
        # type: (int) -> Configuration
        self.wait_before_screenshots = wait_before_screenshots
        return self

    def set_stitch_mode(self, stitch_mode):
        # type: (StitchMode) -> Configuration
        self.stitch_mode = stitch_mode
        return self

    def set_hide_scrollbars(self, hide_scrollbars):
        # type: (bool) -> Configuration
        self.hide_scrollbars = hide_scrollbars
        return self

    def set_hide_caret(self, hide_caret):
        # type: (bool) -> Configuration
        self.hide_caret = hide_caret
        return self

    def set_visual_grid_options(self, *options):
        # type: (*VisualGridOption) -> Configuration
        if options == (None,):
            self.visual_grid_options = None
        else:
            argument_guard.are_(options, VisualGridOption)
            self.visual_grid_options = options
        return self

    def set_disable_browser_fetching(self, disable_browser_fetching):
        # type: (bool) -> Configuration
        self.disable_browser_fetching = disable_browser_fetching
        return self

    def set_enable_cross_origin_rendering(self, enable_cross_origin_rendering):
        # type: (bool) -> Configuration
        self.enable_cross_origin_rendering = enable_cross_origin_rendering
        return self

    def set_dont_use_cookies(self, dont_use_cookies):
        # type: (bool) -> Configuration
        self.dont_use_cookies = dont_use_cookies
        return self

    @overload
    def set_layout_breakpoints(
        self,
        enabled: bool,
        *,
        reload: Optional[bool] = None,
        height_breakpoints: Optional[bool] = None
    ) -> "Configuration":
        pass

    @overload
    def set_layout_breakpoints(self, *breakpoints, **kwargs):
        # type: (*int, Optional[bool]) -> Configuration
        pass

    def set_layout_breakpoints(self, enabled_or_first, *rest, **kwargs):
        reload = kwargs.pop("reload", None)
        height_breakpoints = kwargs.pop("height_breakpoints", None)
        assert not kwargs, "Unexpected keyword arguments {}".format(kwargs)
        if isinstance(enabled_or_first, bool):
            assert (
                not rest
            ), "`reload` and `height_breakpoints` are keyword-only arguments"
            self.layout_breakpoints = LayoutBreakpointsOptions(
                enabled_or_first, reload, height_breakpoints
            )
        elif isinstance(enabled_or_first, int):
            self.layout_breakpoints = LayoutBreakpointsOptions(
                [enabled_or_first] + list(rest), reload, height_breakpoints
            )
        else:
            raise TypeError(
                "{} is not an instance of bool or int".format(enabled_or_first)
            )
        return self

    @overload  # noqa
    def add_browser(self, desktop_browser_info):
        # type: (DesktopBrowserInfo) -> Configuration
        pass

    @overload  # noqa
    def add_browser(self, ios_device_info):
        # type: (IosDeviceInfo) -> Configuration
        pass

    @overload  # noqa
    def add_browser(self, android_device_info):
        # type: (AndroidDeviceInfo) -> Configuration
        pass

    @overload  # noqa
    def add_browser(self, chrome_emulation_info):
        # type: (ChromeEmulationInfo) -> Configuration
        pass

    @overload  # noqa
    def add_browser(self, width, height, browser_type):
        # type: (int, int, BrowserType) -> Configuration
        pass

    @overload  # noqa
    def add_browser(self, width, height, browser_type, baseline_env_name):
        # type: (int, int, BrowserType, Text) -> Configuration
        pass

    def add_browser(self, *args):  # noqa
        if isinstance(args[0], IRenderBrowserInfo):
            self.browsers_info.append(args[0])
        elif (
            isinstance(args[0], int)
            and isinstance(args[1], int)
            and isinstance(args[2], BrowserType)
        ):
            if len(args) == 4:
                baseline_env_name = args[3]
            else:
                baseline_env_name = self.baseline_env_name
            self.browsers_info.append(
                DesktopBrowserInfo(args[0], args[1], args[2], baseline_env_name)
            )
        else:
            raise TypeError(
                "Unsupported parameter: \n\ttype: {} \n\tvalue: {}".format(
                    type(args), args
                )
            )
        return self

    @overload
    def add_browsers(self, renders_info):
        # type:(List[Union[DesktopBrowserInfo,IosDeviceInfo,AndroidDeviceInfo,ChromeEmulationInfo]])->Configuration  # noqa
        pass

    @overload
    def add_browsers(self, renders_info):
        # type:(*Union[DesktopBrowserInfo,IosDeviceInfo,AndroidDeviceInfo,ChromeEmulationInfo])->Configuration
        pass

    def add_browsers(self, *renders_info):
        if len(renders_info) == 1 and is_list_or_tuple(renders_info[0]):
            renders_info = renders_info[0]

        for render_info in renders_info:
            try:
                self.add_browser(render_info)
            except TypeError as e:
                raise TypeError("Wrong argument in .add_browsers()") from e
        return self

    def add_device_emulation(self, device_name, orientation=ScreenOrientation.PORTRAIT):
        # type: (DeviceName, ScreenOrientation) -> Configuration
        argument_guard.not_none(device_name)
        self.add_browser(ChromeEmulationInfo(device_name, orientation))
        return self

    def add_mobile_device(self, mobile_device_info):
        # type: (Union[IosDeviceInfo, AndroidDeviceInfo]) -> Configuration
        return self.add_mobile_devices(mobile_device_info)

    def add_mobile_devices(self, *mobile_device_infos):
        # type: (Union[IosDeviceInfo, AndroidDeviceInfo]) -> Configuration
        return self.add_browsers(*mobile_device_infos)

    def clone(self):
        # type: () -> Configuration
        # TODO: Remove this huck when get rid of Python2
        conf = super(Configuration, self).clone()
        conf.browsers_info = deepcopy(self.browsers_info)
        conf.visual_grid_options = deepcopy(self.visual_grid_options)
        return conf

    def add_multi_device_target(self, *device_targets):
        # type: (*Union[IosMultiDeviceTarget, AndroidMultiDeviceTarget]) -> Configuration
        for device_target in device_targets:
            if isinstance(
                device_target, (IosMultiDeviceTarget, AndroidMultiDeviceTarget)
            ):
                device_target = device_target()

            if isinstance(device_target, IosDeviceTarget):
                ios_device_info = IosDeviceInfo(
                    device_name=device_target.device_name,
                    screen_orientation=device_target.screen_orientation,
                    ios_version=device_target.version,
                )
                self.add_browser(ios_device_info)
            elif isinstance(device_target, AndroidDeviceTarget):
                android_device_info = AndroidDeviceInfo(
                    device_name=device_target.device_name,
                    screen_orientation=device_target.screen_orientation,
                    android_version=device_target.version,
                )
                self.add_browser(android_device_info)
            else:
                raise TypeError(
                    "All arguments must be instances of IosMultiDeviceTarget or AndroidMultiDeviceTarget"
                )
        return self

    def set_mobile_options(self, mobile_options):
        # type: (MobileOptions) -> Configuration
        self.mobile_options = mobile_options
        return self



from __future__ import absolute_import, division, print_function

from .config import Configuration
from .misc import BrowserType, MobileOptions, StitchMode

__all__ = ("StitchMode", "BrowserType", "Configuration", "MobileOptions")



from __future__ import absolute_import, division, print_function, unicode_literals

from applitools.common import deprecated
from applitools.common.fluent.check_settings import CheckSettings, CheckSettingsValues
from applitools.common.fluent.region import (
    AccessibilityRegionByRectangle,
    FloatingRegionByRectangle,
    GetAccessibilityRegion,
    GetFloatingRegion,
    GetRegion,
    RegionByRectangle,
)

deprecated.module(__name__)
__all__ = (
    "AccessibilityRegionByRectangle",
    "CheckSettings",
    "CheckSettingsValues",
    "FloatingRegionByRectangle",
    "GetAccessibilityRegion",
    "GetFloatingRegion",
    "GetRegion",
    "RegionByRectangle",
)



from __future__ import absolute_import, division, print_function

import math
from enum import Enum
from typing import TYPE_CHECKING, Dict, List, Optional, Text, Tuple, Union, overload

import attr

from ._dynamic_regions import DynamicSettings, DynamicTextTypeOrPattern
from .accessibility import AccessibilityRegionType
from .utils import argument_guard
from .utils.converters import round_converter

if TYPE_CHECKING:
    from .ultrafastgrid.render_browser_info import IRenderBrowserinfo
    from .utils.custom_types import CodedRegionPadding, Num

__all__ = (
    "Point",
    "Region",
    "CoordinatesType",
    "RectangleSize",
    "SubregionForStitching",
    "AccessibilityRegion",
    "DynamicRegion",
)


def dx_and_dy(location_or_dx, dy):
    # type: (Union[Point, int], Optional[int]) -> Tuple[int,int]
    dx = location_or_dx
    if dy is None:
        dx, dy = location_or_dx
    return dx, dy


class CoordinatesType(Enum):
    """Encapsulates the type of coordinates used by the region provider."""

    # The coordinates should be used "as is" on the screenshot image.
    # Regardless of the current context.
    SCREENSHOT_AS_IS = "SCREENSHOT_AS_IS"

    # The coordinates should be used "as is" within the current context. For
    # example, if we're inside a frame, the coordinates are "as is",
    # but within the current frame's viewport.
    CONTEXT_AS_IS = "CONTEXT_AS_IS"

    # Coordinates are relative to the context. For example, if we are in
    # a context of a frame in a web page, then the coordinates are relative to
    # the  frame. In this case, if we want to crop an image region based on
    # an element's region, we will need to calculate their respective "as
    # is" coordinates.
    CONTEXT_RELATIVE = "CONTEXT_RELATIVE"


class DictAccessMixin(object):
    """Make dict-like object from attrs class"""

    def __getitem__(self, item):
        fields = [a.name for a in attr.fields(self.__class__)]
        if isinstance(item, int):
            item = fields[item]
        if item not in fields:
            raise KeyError("item: {}, fields: {}".format(item, fields))
        return getattr(self, item)


@attr.s(slots=True, eq=False, hash=True, init=False)
class RectangleSize(DictAccessMixin):
    """Represents a 2D size"""

    width = attr.ib()  # type: int
    height = attr.ib()  # type:int

    def __init__(self, width, height):
        # type: (Num, Num) -> None
        self.width = round_converter(width)  # type: int
        self.height = round_converter(height)  # type: int

    def __str__(self):
        return "RectangleSize({width} x {height})".format(
            width=self.width, height=self.height
        )

    def __eq__(self, other):
        # type: (Union[RectangleSize, Dict]) -> bool
        return self.width == other["width"] and self.height == other["height"]

    def __ne__(self, other):
        # type: (Union[RectangleSize, Dict]) -> bool
        return self.width != other["width"] or self.height != other["height"]

    def __add__(self, other):
        # type: (Union[RectangleSize, Dict]) -> RectangleSize
        return RectangleSize(self.width + other["width"], self.height + other["height"])

    def __radd__(self, other):
        # type: (Union[RectangleSize, Dict]) -> RectangleSize
        return RectangleSize(self.width + other["width"], self.height + other["height"])

    def __sub__(self, other):
        # type: (Union[RectangleSize, Dict]) -> RectangleSize
        return RectangleSize(self.width - other["width"], self.height - other["height"])

    def __rsub__(self, other):
        # type: (Union[RectangleSize, Dict]) -> RectangleSize
        return RectangleSize(other["width"] - self.width, other["height"] - self.height)

    def scale(self, scale_ratio):
        # type: (float) -> RectangleSize
        """Get a scaled version of the current size.

        Args:
            scale_ratio: The ratio by which to scale.

        Returns:
            A scaled version of the current size.
        """
        return RectangleSize(
            round(self.width * scale_ratio), round(self.height * scale_ratio)
        )

    @classmethod
    def from_(cls, obj):
        # type: (Union[dict,IRenderBrowserinfo,RectangleSize])->RectangleSize
        """Creates a new RectangleSize instance."""
        if isinstance(obj, dict):
            return cls(width=obj["width"], height=obj["height"])
        return cls(width=obj.width, height=obj.height)


@attr.s(slots=True, eq=False, init=False)
class Point(DictAccessMixin):
    """A location in a two-dimensional plane."""

    x = attr.ib()  # type: int
    y = attr.ib()  # type: int

    def __init__(self, x, y):
        # type: (int, int) -> None
        self.x = round_converter(x)
        self.y = round_converter(y)

    def __str__(self):
        return "Point({x} x {y})".format(x=self.x, y=self.y)

    def __add__(self, other):
        # type: (Union[Point, int, float]) -> Point
        if isinstance(other, int) or isinstance(other, float):
            return Point(self.x + round(other), self.y + round(other))
        return Point(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        # type: (Union[Point, int, float]) -> Point
        if isinstance(other, int) or isinstance(other, float):
            return Point(self.x - round(other), self.y - round(other))
        return Point(self.x - other.x, self.y - other.y)

    def __neg__(self):
        # type: () -> Point
        return Point(-self.x, -self.y)

    def __mul__(self, scalar):
        # type: (int) -> Point
        return Point(self.x * scalar, self.y * scalar)

    def __div__(self, scalar):
        # type: (int) -> Point
        return Point(round_converter(self.x / scalar), round_converter(self.y / scalar))

    def __eq__(self, other):
        # type: (Union[RectangleSize, Dict]) -> bool
        return self.x == other["x"] and self.y == other["y"]

    @classmethod  # noqa
    def ZERO(cls):
        # type: () -> Point
        return cls(0, 0)

    @classmethod
    def from_(cls, obj):
        # type: (Union[dict, Point]) -> Point
        """Creates a new Point instance."""
        return cls(obj["x"], obj["y"])

    def clone(self):
        # type: () -> Point
        """Make a full copy of this Point.

        Returns:
            New cloned Point instance
        """
        return Point(self.x, self.y)

    @overload  # noqa
    def offset(self, location):
        # type: (Point) -> Point
        pass

    @overload  # noqa
    def offset(self, dx, dy):
        # type: (int, int) -> Point
        pass

    def offset(self, location_or_dx, dy=None):  # noqa
        # type: (Union[Point, int], Optional[int]) -> Point
        """Get a location translated by the specified amount.

        Args:
            location_or_dx: Full amount to offset or just x-coordinate
            dy: The amount to offset the y-coordinate.

        Returns:
            A location translated by the specified amount.
        """
        dx, dy = dx_and_dy(location_or_dx, dy)
        return Point(self.x + dx, self.y + dy)

    def scale(self, scale_ratio):
        # type: (float) -> Point
        """Get a scaled location.

        Args:
            scale_ratio: The ratio by which to scale the results.

        Returns:
            A scaled copy of the current location.
        """
        return Point(
            int(math.ceil(self.x * scale_ratio)), int(math.ceil(self.y * scale_ratio))
        )


@attr.s(slots=True, init=False)
class Rectangle(DictAccessMixin):
    left = attr.ib()  # type: int
    top = attr.ib()  # type: int
    width = attr.ib()  # type: int
    height = attr.ib()  # type: int

    def __init__(
        self,
        left,  # type: Num
        top,  # type: Num
        width,  # type: Num
        height,  # type: Num
    ):
        # type: (...) -> None
        self.left = round_converter(left)  # type: int
        self.top = round_converter(top)  # type: int
        self.width = round_converter(width)  # type: int
        self.height = round_converter(height)  # type: int

    @classmethod  # noqa
    def EMPTY(cls):
        return cls(0, 0, 0, 0)

    @property
    def x(self):
        # type: () -> int
        return self.left

    @property
    def y(self):
        # type: () -> int
        return self.top

    @property
    def right(self):
        # type: () -> int
        return self.left + self.width

    @property
    def bottom(self):
        # type: () -> int
        return self.top + self.height

    @property
    def location(self):
        # type: () -> Point
        """Return the top-left corner as a Point."""
        return Point(self.left, self.top)

    @location.setter
    def location(self, point):
        # type: (Point) -> None
        """Sets the top left corner of the region"""
        argument_guard.not_none(point)
        self.left, self.top = point.x, point.y

    @property
    def area(self):
        return self.width * self.height

    @property
    def bottom_right(self):
        # type: () -> Point
        """
        Returns:
            The bottom-right corner as a Point.
        """
        return Point(self.right, self.bottom)

    @property
    def size(self):
        # type: () -> RectangleSize
        """
        Returns:
            The size of the region.
        """
        return RectangleSize(width=self.width, height=self.height)

    def make_empty(self):
        """Sets the current instance as an empty instance"""
        self.left = self.top = self.width = self.height = 0

    def clip_negative_location(self):
        """Sets the left/top values to 0 if the value is negative"""
        self.left = max(self.left, 0)
        self.top = max(self.top, 0)

    @property
    def is_size_empty(self):
        # type: () -> bool
        """
        Returns:
            true if the region's size is 0, false otherwise.
        """
        return self.width <= 0 or self.height <= 0

    @property
    def is_empty(self):
        # type: () -> bool
        """Checks whether the rectangle is empty.

        Returns:
            True if the rectangle is empty. Otherwise False.
        """
        return self.left == self.top == self.width == self.height == 0

    @property
    def middle_offset(self):
        # type: () -> Point
        return Point(int(round(self.width / 2)), int(round(self.height / 2)))

    @overload  # noqa
    def offset(self, location):
        # type: (Point) -> Rectangle
        pass

    @overload  # noqa
    def offset(self, dx, dy):
        # type: (int, int) -> Rectangle
        pass

    def offset(self, location_or_dx, dy=None):  # noqa
        # type: (Union[Point, int], Optional[int]) -> Rectangle
        """Get an offset region.

        Args:
            location_or_dx: Full amount to offset or just x-coordinate
            dy: The amount to offset the y-coordinate.

        Returns:
            A region with an offset location.
        """
        dx, dy = dx_and_dy(location_or_dx, dy)
        location = self.location.offset(dx, dy)
        return Rectangle(
            left=location.x,
            top=location.y,
            width=self.size["width"],
            height=self.size["height"],
        )

    def get_sub_regions(
        self,
        max_size,  # type: RectangleSize
        overlap_and_crop,  # type: int
        l2p_scale_ratio,  # type: float
        physical_rect_in_screenshot,  # type: Region
    ):
        # type: (...) -> List[SubregionForStitching]
        overlap = crop = overlap_and_crop
        tiles = overlapping_tiles_from_rectangle(self, max_size, overlap + crop)
        return [
            SubregionForStitching.from_tile(
                tile,
                max_size,
                self,
                l2p_scale_ratio,
                crop,
                physical_rect_in_screenshot,
            )
            for tile in tiles
        ]


@attr.s(slots=True, init=False)
class AccessibilityRegion(Rectangle):
    """An accessibility region identified by left,top, width, height and type"""

    left = attr.ib()  # type: int
    top = attr.ib()  # type: int
    width = attr.ib()  # type: int
    height = attr.ib()  # type: int
    type = attr.ib(
        type=AccessibilityRegionType,
    )  # type: AccessibilityRegionType

    def __init__(
        self,
        left,  # type: Num
        top,  # type: Num
        width,  # type: Num
        height,  # type: Num
        type,  # type: AccessibilityRegionType
    ):
        # type: (...) -> None
        super(AccessibilityRegion, self).__init__(left, top, width, height)
        if isinstance(type, str):
            if type == "None":
                self.type = None
                return
        self.type = AccessibilityRegionType(type)

    @overload  # noqa
    def from_(self, accessibility_region):
        # type: (Union[Dict, AccessibilityRegion]) -> AccessibilityRegion
        pass

    @overload  # noqa
    def from_(self, region, accessibility_type):
        # type: (Union[Region,Rectangle],AccessibilityRegionType)->AccessibilityRegion
        pass

    @classmethod  # noqa
    def from_(cls, obj, obj2=None):
        # type: (...)->AccessibilityRegion
        """Creates a new Region instance."""
        if isinstance(obj, AccessibilityRegion):
            return cls(obj.left, obj.top, obj.width, obj.height, obj.type)
        elif (
            isinstance(obj, Region)
            or isinstance(obj, Rectangle)
            and isinstance(obj2, AccessibilityRegionType)
        ):
            return cls(obj.left, obj.top, obj.width, obj.height, obj2)
        elif isinstance(obj, dict):
            return cls(
                obj["left"],
                obj["top"],
                obj["width"],
                obj["height"],
                AccessibilityRegionType(obj["type"]),
            )
        raise ValueError("Wrong parameters passed")

    def offset(self, location_or_dx, dy=None):  # noqa
        # type: (Union[Point, int], Optional[int]) -> AccessibilityRegion
        r = super(AccessibilityRegion, self).offset(location_or_dx, dy)
        return AccessibilityRegion.from_(r, self.type)


@attr.s(slots=True, init=False)
class DynamicRegion(Rectangle):
    """A dynamic region identified by left,top, width, height and type/pattern"""

    left = attr.ib()  # type: int
    top = attr.ib()  # type: int
    width = attr.ib()  # type: int
    height = attr.ib()  # type: int
    dynamic_settings = attr.ib(default=None)  # type: Optional[DynamicSettings]
    padding = attr.ib(default=None)  # type: Optional[CodedRegionPadding]


@attr.s(slots=True, init=False)
class Region(Rectangle):
    """A rectangle identified by left,top, width, height."""

    left = attr.ib()  # type: int
    top = attr.ib()  # type: int
    width = attr.ib()  # type: int
    height = attr.ib()  # type: int
    coordinates_type = attr.ib(
        converter=CoordinatesType,
        type=CoordinatesType,
    )  # type: CoordinatesType

    def __init__(
        self,
        left,  # type: Num
        top,  # type: Num
        width,  # type: Num
        height,  # type: Num
        coordinates_type=CoordinatesType.SCREENSHOT_AS_IS,  # type: CoordinatesType
    ):
        # type: (...) -> None
        super(Region, self).__init__(left, top, width, height)
        self.coordinates_type = CoordinatesType(coordinates_type)

    def __str__(self):
        return "Region({left}, {top}, {width} x {height}, {type})".format(
            left=self.left,
            top=self.top,
            width=self.width,
            height=self.height,
            type=self.coordinates_type.value,
        )

    def padding(self, other):
        # type: (CodedRegionPadding) -> Region
        if isinstance(other, dict):
            return Region(
                left=self.left + other.get("left", 0),
                top=self.top + other.get("top", 0),
                width=self.width + other.get("right", 0),
                height=self.height + other.get("bottom", 0),
                coordinates_type=self.coordinates_type,
            )
        else:
            raise TypeError("Unsupported flow")

    @classmethod  # noqa
    @overload
    def from_(cls, location, size):
        # type: (Union[Point, Dict], Union[Dict,RectangleSize]) -> Region
        pass

    @classmethod  # noqa
    @overload
    def from_(cls, region):
        # type: (Union[Region, Dict]) -> Region
        pass

    @classmethod  # noqa
    @overload
    def from_(cls, rectangle, coordinates_type):
        # type: (Rectangle, Optional[CoordinatesType]) -> Region
        pass

    @classmethod  # noqa
    def from_(cls, obj, obj2=None):
        # type:( ... )->Region
        """Creates a new Region instance."""
        if isinstance(obj, Rectangle):
            if isinstance(obj2, CoordinatesType):
                return cls(obj.left, obj.top, obj.width, obj.height, obj2)
            return cls(obj.left, obj.top, obj.width, obj.height)
        if isinstance(obj, Region):
            return cls(obj.left, obj.top, obj.width, obj.height, obj.coordinates_type)
        elif isinstance(obj, Point) or isinstance(obj, dict) and obj2:
            return cls(obj["x"], obj["y"], obj2["width"], obj2["height"])
        elif isinstance(obj, dict) and obj2 is None:
            return cls(
                obj["left"],
                obj["top"],
                obj["width"],
                obj["height"],
            )
        raise ValueError("Wrong parameters passed")

    def clone(self):
        # type: () -> Region
        """
        Returns:
            The cloned instance of Region.
        """
        return Region(
            self.left, self.top, self.width, self.height, self.coordinates_type
        )

    def contains(self, other):
        # type: (Union[Point, Region]) -> bool
        """Return true if a point is inside the rectangle.

        Args:
            other: element for check

        Returns:
            True if the point is inside the rectangle. Otherwise False.
        """
        if isinstance(other, Point):
            x, y = other
            return self.left <= x <= self.right and self.top <= y <= self.bottom  # noqa
        else:
            right = self.left + self.width
            pt_right = other.left + other.width

            bottom = self.top + self.height
            pt_bottom = other.top + other.height

            return (
                self.top <= other.top
                and self.left <= other.left
                and bottom >= pt_bottom
                and right >= pt_right
            )

    def overlaps(self, other):
        # type: (Region) -> bool
        """Return true if a rectangle overlaps this rectangle.

        Args:
            other:
        """
        return (
            self.left <= other.left <= self.right
            or other.left <= self.left <= other.right
        ) and (
            self.top <= other.top <= self.bottom
            or other.top <= self.top <= other.bottom
        )

    def is_intersected(self, other):
        return self.overlaps(other)

    def intersect(self, other):
        # type: (Region) -> Region
        # If the regions don't overlap, the intersection is empty
        """
        Args:
            other:
        """
        if not self.overlaps(other):
            return Region.EMPTY()
        intersection_left = self.left if self.left >= other.left else other.left
        intersection_top = self.top if self.top >= other.top else other.top
        intersection_right = self.right if self.right <= other.right else other.right
        intersection_bottom = (
            self.bottom if self.bottom <= other.bottom else other.bottom
        )
        left, top = intersection_left, intersection_top
        width = intersection_right - intersection_left
        height = intersection_bottom - intersection_top
        return Region(left, top, width, height, self.coordinates_type)

    def offset(self, location_or_dx, dy=None):  # noqa
        # type: (Union[Point, int], Optional[int]) -> Region
        r = super(Region, self).offset(location_or_dx, dy)
        return Region.from_(r, self.coordinates_type)

    def scale(self, scale_ratio):
        # type: (float) -> Region
        """Get a region which is a scaled version of the current region.
        IMPORTANT: This also scales the LOCATION(!!) of the region (not just its
        size).

        Args:
            scale_ratio: The ratio by which to scale the region.

        Returns:
            A new region which is a scaled version of the current region.
        """
        return Region(
            left=int(math.ceil(self.left * scale_ratio)),
            top=int(math.ceil(self.top * scale_ratio)),
            width=int(math.ceil(self.width * scale_ratio)),
            height=int(math.ceil(self.height * scale_ratio)),
            coordinates_type=self.coordinates_type,
        )


@attr.s
class SubregionForStitching(object):
    scroll_to = attr.ib()  # type: Point
    paste_physical_location = attr.ib()  # type: Point
    physical_crop_area = attr.ib()  # type: Region
    logical_crop_area = attr.ib()  # type: Region

    @classmethod
    def from_tile(
        cls,
        tile,  # type: Rectangle
        scroll_size,  # type: RectangleSize
        region,  # type: Rectangle
        l2p_scale_ratio,  # type: float
        crop_size,  # type: int
        rect_in_screenshot,  # type: Region
    ):
        # type: (...) -> SubregionForStitching
        # It's not possible to scroll too far leaving less than scroll_size area
        # visible, scroll as far possible then
        scroll_to = Point(
            min(tile.left, region.right - scroll_size.width),
            min(tile.top, region.bottom - scroll_size.height),
        )
        # If we couldn't scroll to the tile exactly, redundant pixels should be cropped
        scroll_crop = RectangleSize(
            max(0, tile.left - scroll_to.x), max(0, tile.top - scroll_to.y)
        )
        physical_crop_area = Region(
            round(rect_in_screenshot.left + scroll_crop.width * l2p_scale_ratio),
            round(rect_in_screenshot.top + scroll_crop.height * l2p_scale_ratio),
            round(tile.width * l2p_scale_ratio),
            round(tile.height * l2p_scale_ratio),
            rect_in_screenshot.coordinates_type,
        )
        # Don't do additional "logical" cropping on top and left borders
        crop = Point(
            0 if tile.left == region.left else crop_size,
            0 if tile.top == region.top else crop_size,
        )
        logical_crop_area = Region(
            crop.x, crop.y, tile.width - crop.x, tile.height - crop.y
        )
        paste_physical_location = tile.location + crop - region.location
        return cls(
            scroll_to, paste_physical_location, physical_crop_area, logical_crop_area
        )


def tiles_from_rectangle(rect, tile_size):
    # type: (Rectangle, RectangleSize) -> List[Rectangle]
    """Breaks the rect rectangle into tile_size sized tiles
    Smaller pieces are placed near bottom and right borders if rect doesn't split even.
    """
    # Having smaller tiles close to top and left borders could have saved from
    # calculations of scrolling and cropping offsets for them later, but that breaks
    # existing tests that leave scrollbar visible when stitching
    tiles = []
    for top in range(rect.top, rect.bottom, tile_size.height):
        height = min(tile_size.height, rect.bottom - top)
        for left in range(rect.left, rect.right, tile_size.width):
            width = min(tile_size.width, rect.right - left)
            tiles.append(Rectangle(left, top, width, height))
    return tiles  # noqa


def overlapping_tiles_from_rectangle(rect, tile_size, overlap):
    # type: (Rectangle, RectangleSize, int) -> List[Rectangle]
    """Breaks the rect rectangle into tile_size sized tiles where most tiles have
    redundant space on their right and bottom edge of the `overlap` size,
    it is covered by the next tile to it's right and below.
    Tiles on the right and bottom edge do not have that space to be covered."""
    overlap_reduced_rect = Rectangle(
        rect.left, rect.top, rect.width - overlap, rect.height - overlap
    )
    overlap_reduced_tile = tile_size - RectangleSize(overlap, overlap)
    tiles = tiles_from_rectangle(overlap_reduced_rect, overlap_reduced_tile)
    return [
        Rectangle(t.left, t.top, t.width + overlap, t.height + overlap) for t in tiles
    ]



from __future__ import absolute_import, division, print_function

from enum import Enum

import attr

__all__ = ("SessionType", "FailureReports", "ServerInfo")


@attr.s
class ServerInfo(object):
    logs_dir = attr.ib()


class SessionType(Enum):
    # default type of sessions.
    SEQUENTIAL = "SEQUENTIAL"
    # a timing test session
    PROGRESSION = "PROGRESSION"


class FailureReports(Enum):
    """
    Failures are either reported immediately when they are detected, or when the test is closed.
    """

    # Failures are reported immediately when they are detected.
    IMMEDIATE = "Immediate"
    # Failures are reported when tests are completed (i.e., when
    # :py:class:`EyesBase.close()` is called).
    ON_CLOSE = "OnClose"



from __future__ import absolute_import, division, print_function

from base64 import b64encode
from io import BytesIO
from typing import TYPE_CHECKING

import attr

from ..optional_deps import Image, PathLike, fspath
from .web_check_settings import WebCheckSettings, WebCheckSettingsValues

if TYPE_CHECKING:
    from typing import ByteString, Optional, Text, Union


@attr.s
class ImagesCheckSettingsValues(WebCheckSettingsValues):
    image = attr.ib(default=None)  # type: Optional[Text]
    dom = attr.ib(default=None)  # type: Optional[Text]


@attr.s
class ImagesCheckSettings(WebCheckSettings):
    Values = ImagesCheckSettingsValues

    def with_dom(self, dom):
        # type: (Text) -> ImagesCheckSettings
        """
        Attach given DOM source text to the image.
        Needed by the Root Cause Analysis tool
        """
        self.values.dom = dom
        return self


def image_path_or_bytes(image_or_path):
    # type: (Union[ByteString, Image, Text, PathLike]) -> Text
    if isinstance(image_or_path, bytes):
        image_bytes = b64encode(image_or_path)
        return image_bytes.decode("utf-8")
    elif isinstance(image_or_path, PathLike):
        return fspath(image_or_path)
    elif isinstance(image_or_path, Image):
        image_bytes = BytesIO()
        image_or_path.save(image_bytes, format="PNG")
        image_bytes = b64encode(image_bytes.getvalue())
        return image_bytes.decode("utf-8")
    elif isinstance(image_or_path, str):
        return image_or_path
    else:
        raise ValueError("Invalid image type", type(image_or_path))



# GENERATED FILE #
from enum import Enum

from applitools.common.utils.general_utils import DeprecatedEnumVariant

__all__ = ("IosDeviceName",)


class IosDeviceName(Enum):
    iPad_7 = "iPad (7th generation)"
    iPad_9 = "iPad (9th generation)"
    iPad_10 = "iPad (10th generation)"
    iPad_mini_6 = "iPad mini (6th generation)"
    iPad_Air_4 = "iPad Air (4th generation)"
    iPad_Air_11_inch_M3 = "iPad Air 11-inch (M3)"
    iPad_Air_13_inch_M3 = "iPad Air 13-inch (M3)"
    iPad_A16 = "iPad (A16)"
    iPad_Pro_11_inch_4 = "iPad Pro (11-inch) (4th generation)"
    iPad_Pro_11_inch_M5 = "iPad Pro 11-inch (M5)"
    iPad_Pro_13_inch_M5 = "iPad Pro 13-inch (M5)"
    iPad_Pro_12_9_inch_3 = "iPad Pro (12.9-inch) (3rd generation)"
    iPhone_SE_2 = "iPhone SE (2nd generation)"
    iPhone_SE_3 = "iPhone SE (3rd generation)"
    iPhone_XR = "iPhone XR"
    iPhone_Xs = "iPhone Xs"
    iPhone_11 = "iPhone 11"
    iPhone_11_Pro_Max = "iPhone 11 Pro Max"
    iPhone_11_Pro = "iPhone 11 Pro"
    iPhone_12 = "iPhone 12"
    iPhone_12_mini = "iPhone 12 mini"
    iPhone_12_Pro_Max = "iPhone 12 Pro Max"
    iPhone_12_Pro = "iPhone 12 Pro"
    iPhone_13 = "iPhone 13"
    iPhone_13_mini = "iPhone 13 mini"
    iPhone_13_Pro_Max = "iPhone 13 Pro Max"
    iPhone_13_Pro = "iPhone 13 Pro"
    iPhone_14 = "iPhone 14"
    iPhone_14_Pro_Max = "iPhone 14 Pro Max"
    iPhone_14_Pro = "iPhone 14 Pro"
    iPhone_14_Plus = "iPhone 14 Plus"
    iPhone_15 = "iPhone 15"
    iPhone_15_Pro_Max = "iPhone 15 Pro Max"
    iPhone_15_Pro = "iPhone 15 Pro"
    iPhone_15_Plus = "iPhone 15 Plus"
    iPhone_16 = "iPhone 16"
    iPhone_16_Pro_Max = "iPhone 16 Pro Max"
    iPhone_16_Pro = "iPhone 16 Pro"
    iPhone_16_Plus = "iPhone 16 Plus"
    iPhone_17 = "iPhone 17"
    iPhone_17_Pro = "iPhone 17 Pro"
    iPhone_17_Pro_Max = "iPhone 17 Pro Max"
    iPhone_Air = "iPhone Air"

    @DeprecatedEnumVariant
    def iPad_Pro_3(self):
        """`iPad_Pro_3` is deprecated. Use `iPad_Pro_12_9_inch_3` instead"""
        return self.iPad_Pro_12_9_inch_3

    @DeprecatedEnumVariant
    def iPad_Pro_4(self):
        """`iPad_Pro_4` is deprecated. Use `iPad_Pro_11_inch_4` instead"""
        return self.iPad_Pro_11_inch_4

    @DeprecatedEnumVariant
    def iPhone_XS(self):
        """`iPhone_XS` is deprecated. Use `iPhone_Xs` instead"""
        return self.iPhone_Xs



from __future__ import absolute_import, division, print_function

from typing import TYPE_CHECKING

from .__version__ import __version__
from .command_context import CommandContext
from .object_registry import ObjectRegistry

if TYPE_CHECKING:
    from typing import List, Optional, Union

    from .connection import USDKConnection


class USDKProtocol(object):
    _CommandContext = CommandContext
    _ObjectRegistry = ObjectRegistry
    SDK_INFO = "eyes-common", __version__

    @classmethod
    def object_registry(cls):
        return cls._ObjectRegistry()

    @classmethod
    def context(cls, connection, object_registry=None):
        # type: (USDKConnection, Optional[ObjectRegistry]) -> CommandContext
        return cls._CommandContext(connection, object_registry or cls.object_registry())

    @classmethod
    def commands_or_kind(cls):
        # type: () -> Union[str, List[str]]
        return cls._CommandContext.commands_or_kind()



from __future__ import absolute_import, division, print_function

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional, Text, Union

    from ..object_registry import ObjectRegistry
    from ..optional_deps import WebElement


BY_CSS_SELECTOR = "css selector"


class PathNodeValue(object):
    def __eq__(self, other):
        # type: (PathNodeValue) -> bool
        return type(self) is type(other) and vars(self) == vars(other)

    def __ne__(self, other):
        # type: (PathNodeValue) -> bool
        return not self == other


class ElementReference(PathNodeValue):
    def __init__(self, element):
        # type: (WebElement) -> None
        self.element = element

    def _repr_args(self, _):
        # type: (bool) -> Text
        return repr(self.element)

    def _to_dict(self, registry):
        # type: (ObjectRegistry) -> dict
        return registry.marshal_element(self.element)


class ElementSelector(PathNodeValue):
    def __init__(self, selector_or_by, selector=None):
        # type: (Text, Optional[Text]) -> None
        if selector is None:
            self.by = BY_CSS_SELECTOR
            self.selector = selector_or_by
        else:
            self.by = selector_or_by
            self.selector = selector

    def _repr_args(self, skip_default_css_selector_type):
        # type: (bool) -> Text
        if self.by == BY_CSS_SELECTOR and skip_default_css_selector_type:
            return repr(self.selector)
        else:
            return "{}, {!r}".format(_by_repr(self.by), self.selector)

    def _to_dict(self, _):
        # type: (ObjectRegistry) -> dict
        return {"type": self.by, "selector": self.selector}


class FrameSelector(PathNodeValue):
    def __init__(self, number_or_id_or_name):
        # type: (Union[int, Text]) -> None
        self.number_or_id_or_name = number_or_id_or_name

    def _repr_args(self, _):
        # type: (bool) -> Text
        return repr(self.number_or_id_or_name)

    def _to_dict(self, _):
        # type: (ObjectRegistry) -> dict
        return {"selector": self.number_or_id_or_name}


class TargetPathLocator(object):
    def __init__(self, parent, value):
        # type: (Optional[TargetPathLocator], PathNodeValue) -> None
        self.parent = parent
        self.value = value

    def to_dict(self, registry):
        # type: (ObjectRegistry) -> dict
        converted = self.value._to_dict(registry)  # noqa
        parent = self.parent
        while parent:
            converted = {parent.FACTORY_METHOD: converted}
            converted.update(parent.value._to_dict(registry))  # noqa
            parent = parent.parent
        return converted

    def _repr(self, skip_default_css_selector_type):
        # type: (bool) -> Text
        parent = repr(self.parent) if self.parent else "TargetPath"
        args = self.value._repr_args(skip_default_css_selector_type)  # noqa
        return parent + ".{}({})".format(self.FACTORY_METHOD, args)

    def __eq__(self, other):
        # type: (TargetPathLocator) -> bool
        return type(self) is type(other) and vars(self) == vars(other)

    def __ne__(self, other):
        # type: (TargetPathLocator) -> bool
        return not self == other

    def __repr__(self):
        # type: () -> Text
        return self._repr(True)


class TargetPath(object):
    @staticmethod
    def region(element_or_selector_or_by, selector=None):
        return _region(None, element_or_selector_or_by, selector)

    @staticmethod
    def frame(element_or_number_or_id_or_name_or_by, selector=None):
        return _frame(None, element_or_number_or_id_or_name_or_by, selector)

    @staticmethod
    def shadow(element_or_selector_or_by, selector=None):
        return _shadow(None, element_or_selector_or_by, selector)


class ShadowDomLocator(TargetPathLocator):
    FACTORY_METHOD = "shadow"

    def region(self, element_or_selector_or_by, selector=None):
        return _region(self, element_or_selector_or_by, selector)

    def frame(self, element_or_number_or_id_or_name_or_by, selector=None):
        return _frame(self, element_or_number_or_id_or_name_or_by, selector)

    def shadow(self, element_or_selector_or_by, selector=None):
        return _shadow(self, element_or_selector_or_by, selector)


class RegionLocator(TargetPathLocator):
    FACTORY_METHOD = "region"


class FrameLocator(TargetPathLocator):
    FACTORY_METHOD = "frame"

    def region(self, element_or_selector_or_by, selector=None):
        return _region(self, element_or_selector_or_by, selector)

    def frame(self, element_or_number_or_id_or_name_or_by, selector=None):
        return _frame(self, element_or_number_or_id_or_name_or_by, selector)

    def shadow(self, element_or_selector_or_by, selector=None):
        return _shadow(self, element_or_selector_or_by, selector)

    def __repr__(self):
        return self._repr(False)


def _frame(parent, element_or_number_or_id_or_name_or_by, selector=None):
    if selector is not None:
        by = element_or_number_or_id_or_name_or_by
        return FrameLocator(parent, ElementSelector(by, selector))
    elif isinstance(element_or_number_or_id_or_name_or_by, (int, str)):
        number_or_id_or_name = element_or_number_or_id_or_name_or_by
        return FrameLocator(parent, FrameSelector(number_or_id_or_name))
    else:
        element = element_or_number_or_id_or_name_or_by
        return FrameLocator(parent, ElementReference(element))


def _shadow(parent, element_or_selector_or_by, selector=None):
    if selector is not None:
        by = element_or_selector_or_by
        return ShadowDomLocator(parent, ElementSelector(by, selector))
    elif isinstance(element_or_selector_or_by, str):
        selector = element_or_selector_or_by
        return ShadowDomLocator(parent, ElementSelector(BY_CSS_SELECTOR, selector))
    else:
        element = element_or_selector_or_by
        return ShadowDomLocator(parent, ElementReference(element))


def _region(parent, element_or_selector_or_by, selector=None):
    if selector is not None:
        by = element_or_selector_or_by
        return RegionLocator(parent, ElementSelector(by, selector))
    elif isinstance(element_or_selector_or_by, str):
        selector = element_or_selector_or_by
        return RegionLocator(parent, ElementSelector(BY_CSS_SELECTOR, selector))
    else:
        element = element_or_selector_or_by
        return RegionLocator(parent, ElementReference(element))


def _generate_by_repr_map():
    try:  # look for appium>=2.1
        from appium.webdriver.common.appiumby import AppiumBy

        enumerations = [AppiumBy]
    except ImportError:
        try:  # look for old appium otherwise
            from appium.webdriver.common.mobileby import MobileBy

            enumerations = [MobileBy]
        except ImportError:  # no appium installed, that's also fine
            enumerations = []
    try:
        from selenium.webdriver.common.by import By

        enumerations.append(By)
    except ImportError:
        pass

    m = {}
    for by in enumerations:
        prefix = by.__name__ + "."
        m.update((v, prefix + k) for k, v in vars(by).items() if not k.startswith("_"))
    return m


def _by_repr(val):
    return _BY_REPR.get(val, repr(val))


_BY_REPR = _generate_by_repr_map()



from __future__ import absolute_import, division, print_function

from collections import namedtuple
from enum import Enum
from typing import TYPE_CHECKING, List, Optional, Text, Tuple

import attr

from .accessibility import SessionAccessibilityStatus
from .geometry import RectangleSize
from .match import ImageMatchSettings
from .ultrafastgrid import IRenderBrowserInfo

if TYPE_CHECKING:
    from . import ProxySettings

__all__ = (
    "TestResults",
    "TestResultsStatus",
    "TestResultsSummary",
    "TestResultContainer",
)

ConnectionConfig = namedtuple("ConnectionConfig", "server_url api_key proxy")


class TestResultsStatus(Enum):
    """
    Status values for tests results.
    """

    Passed = "Passed"
    Unresolved = "Unresolved"
    Failed = "Failed"

    __test__ = False  # avoid warnings in test frameworks


@attr.s
class SessionUrls(object):
    batch = attr.ib(default=None)  # type: Text
    session = attr.ib(default=None)  # type: Text


@attr.s
class StepInfo(object):
    @attr.s
    class AppUrls(object):
        step = attr.ib(default=None)  # type: Text
        step_editor = attr.ib(default=None)  # type: Text

    @attr.s
    class ApiUrls(object):
        baseline_image = attr.ib(default=None)  # type: Text
        current_image = attr.ib(default=None)  # type: Text
        diff_image = attr.ib(default=None)  # type: Text
        checkpoint_image = attr.ib(default=None)  # type: Text
        checkpoint_image_thumbnail = attr.ib(default=None)  # type: Text
        side_by_side_image = attr.ib(default=None)  # type: Text

    name = attr.ib(default=None)  # type: Text
    is_different = attr.ib(default=None)  # type: bool
    has_baseline_image = attr.ib(default=None)  # type: bool
    has_current_image = attr.ib(default=None)  # type: bool
    has_checkpoint_image = attr.ib(default=None)  # type: bool
    api_urls = attr.ib(default=None, type=ApiUrls)  # type: ApiUrls
    app_urls = attr.ib(default=None, type=AppUrls)  # type: AppUrls

    def __getitem__(self, key):
        # type: (str) -> Any
        """Support dictionary-style access to attributes"""
        return getattr(self, key)


@attr.s
class TestResults(object):
    """
    Eyes test results.
    """

    steps = attr.ib(default=0)  # type: int
    matches = attr.ib(default=0)  # type: int
    mismatches = attr.ib(default=0)  # type: int
    missing = attr.ib(default=0)  # type: int
    exact_matches = attr.ib(default=0, repr=False)  # type: int
    strict_matches = attr.ib(default=0, repr=False)  # type: int
    content_matches = attr.ib(default=0, repr=False)  # type: int
    layout_matches = attr.ib(default=0, repr=False)  # type: int
    none_matches = attr.ib(default=0, repr=False)  # type: int
    is_new = attr.ib(default=None, repr=False)  # type: Optional[bool]

    url = attr.ib(default=None)  # type: Optional[Text]
    status = attr.ib(
        converter=lambda v: TestResultsStatus(v) if v else None,
        type=TestResultsStatus,
        default=None,
        repr=False,
    )  # type: Optional[TestResultsStatus]

    name = attr.ib(default=None, repr=False)  # type: Text
    secret_token = attr.ib(default=None, repr=False)  # type: Text
    id = attr.ib(default=None, repr=False)  # type: Text
    app_name = attr.ib(default=None, repr=False)  # type: Text
    batch_name = attr.ib(default=None, repr=False)  # type: Text
    batch_id = attr.ib(default=None, repr=False)  # type: Text
    branch_name = attr.ib(default=None, repr=False)  # type: Text
    host_os = attr.ib(default=None, repr=False)  # type: Text
    host_app = attr.ib(default=None, repr=False)  # type: Text
    host_display_size = attr.ib(
        default=None, repr=False, type=RectangleSize
    )  # type: RectangleSize
    started_at = attr.ib(default=None, repr=False)  # type: Text
    duration = attr.ib(default=None, repr=False)  # type: int
    is_different = attr.ib(default=None, repr=False)  # type: bool
    is_aborted = attr.ib(default=None, repr=False)  # type: bool
    is_empty = attr.ib(default=None, repr=False)  # type: bool
    app_urls = attr.ib(default=None, repr=False, type=SessionUrls)  # type: SessionUrls
    api_urls = attr.ib(default=None, repr=False, type=SessionUrls)  # type: SessionUrls
    steps_info = attr.ib(
        default=None, repr=False, type=StepInfo
    )  # type: List[StepInfo]
    baseline_id = attr.ib(default=None, repr=False)  # type: Text
    default_match_settings = attr.ib(
        default=None, repr=False, type=ImageMatchSettings
    )  # type: ImageMatchSettings
    accessibility_status = attr.ib(
        default=None, type=SessionAccessibilityStatus
    )  # type: SessionAccessibilityStatus
    user_test_id = attr.ib(default=None)  # type: Optional[Text]
    _connection_config = attr.ib(
        default=(None, None, None), eq=False, order=False, repr=False
    )  # type: Tuple[Optional[Text], Optional[Text], Optional[ProxySettings]]
    __test__ = False  # avoid warnings in test frameworks

    @property
    def is_passed(self):
        # type: () -> bool
        return self.status == TestResultsStatus.Passed

    @property
    def is_unresolved(self):
        # type: () -> bool
        return self.status == TestResultsStatus.Unresolved

    @property
    def is_failed(self):
        # type: () -> bool
        return self.status == TestResultsStatus.Failed

    def set_connection_config(self, server_url, api_key, proxy_settings):
        # type: (Text, Text, Optional[ProxySettings]) -> None
        self._connection_config = ConnectionConfig(server_url, api_key, proxy_settings)

    def delete(self):
        # type: () -> None
        from .command_executor import CommandExecutor
        from .runner import EyesRunner

        cmd = CommandExecutor.get_instance(EyesRunner.Protocol)
        cmd.core_delete_test(self)


@attr.s(repr=False, str=False)
class TestResultContainer(object):
    test_results = attr.ib(default=None, type=TestResults)  # type: TestResults
    browser_info = attr.ib(default=None)  # type: Optional[IRenderBrowserInfo]
    exception = attr.ib(default=None)  # type: Optional[Exception]
    user_test_id = attr.ib(default=None)  # type: Optional[Text]
    __test__ = False  # avoid warnings in test frameworks

    def __str__(self):
        browser_info = (
            "\n browser_info = {}".format(self.browser_info)
            if self.browser_info
            else ""
        )
        return (
            "TestResultContainer{{"
            "\n test_results = {test_results}"
            "{browser_info}"
            "\n exception = {exception}"
            "}}".format(
                test_results=self.test_results,
                browser_info=browser_info,
                exception=self.exception,
            )
        )


@attr.s(repr=False, str=False, eq=False)
class TestResultsSummary(object):
    results = attr.ib(
        factory=list, type=TestResultContainer
    )  # type: List[TestResultContainer]
    exceptions = attr.ib(default=0)  # type: int
    passed = attr.ib(default=0)  # type: int
    unresolved = attr.ib(default=0)  # type: int
    failed = attr.ib(default=0)  # type: int
    mismatches = attr.ib(default=0)  # type: int
    missing = attr.ib(default=0)  # type: int
    matches = attr.ib(default=0)  # type: int
    __test__ = False  # avoid warnings in test frameworks

    @property
    def all_results(self):
        # type: () -> List[TestResultContainer]
        return self.results

    def size(self):
        # type: () -> int
        return len(self)

    def __eq__(self, other):
        return (
            self.passed == other.passed
            and self.unresolved == other.unresolved
            and self.failed == other.failed
            and self.mismatches == other.mismatches
            and self.missing == other.missing
            and self.matches == other.matches
        )

    def __len__(self):
        return len(self.results)

    def __iter__(self):
        return iter(self.results)

    def __getitem__(self, item):
        # type: (int) -> TestResultContainer
        return self.results[item]

    def __str__(self):
        return """
result summary {{
    all results={all_results}
    passed={passed}
    unresolved={unresolved}
    failed={failed}
    exceptions={exceptions}
    mismatches={mismatches}
    missing={missing}
    matches={matches}
}}""".format(
            all_results="\n".join(map(str, self.results)),
            passed=self.passed,
            unresolved=self.unresolved,
            failed=self.failed,
            exceptions=self.exceptions,
            mismatches=self.mismatches,
            missing=self.missing,
            matches=self.matches,
        )



from enum import Enum
from typing import Text, Union

import attr

from applitools.common.utils.general_utils import CustomEnumMethod

__all__ = ("DynamicTextType", "DynamicSettings")


@attr.s(slots=True)
class DynamicSettings:
    ignore_patterns = attr.ib(type=str, factory=list)

    @classmethod
    def from_(cls, dynamic_text_or_patterns):
        obj = cls()
        for pattern in dynamic_text_or_patterns:
            if isinstance(pattern, str):
                obj.ignore_patterns.append(pattern)
            elif isinstance(pattern, DynamicTextType):
                obj.ignore_patterns.append(pattern.value)
        return obj

    def to_dict(self):
        return attr.asdict(self)


class DynamicTextType(Enum):
    TextField = "TextField"
    Number = "Number"
    Date = "Date"
    Link = "Link"
    Email = "Email"
    Currency = "Currency"

    @CustomEnumMethod
    def Custom(cls, pattern):
        # type: (Text) -> Text
        """Set desired pattern"""
        return pattern


DynamicTextTypeOrPattern = Union[DynamicTextType, str]



from __future__ import absolute_import, division, print_function

from typing import TYPE_CHECKING, Dict, List, Text, Union

if TYPE_CHECKING:
    from ..geometry import RectangleSize, Region
    from ..optional_deps import AppiumWebElement, WebElement
    from ..triggers import ActionTrigger

    ViewPort = Union[Dict[Text, int], RectangleSize]  # typedef
    CodedRegionPadding = Union[int, Dict[Text, int]]  # typedef
    Num = Union[int, float]

    AnyWebElement = Union[WebElement, AppiumWebElement]  # typedef

    FrameNameOrId = Text  # typedef
    FrameIndex = int  # typedef
    FrameReference = Union[FrameNameOrId, FrameIndex, AnyWebElement]  # typedef

    # could contain MouseTrigger, TextTrigger
    UserInputs = List[ActionTrigger]  # typedef
    RegionOrElement = Union[AnyWebElement, Region]  # typedef

    SeleniumBy = Text  # typedef
    BySelector = List[SeleniumBy, Text]  # typedef
    CssSelector = Text  # typedef
    REGION_VALUES = Union[Region, CssSelector, AnyWebElement, BySelector]  # typedef
    FLOATING_VALUES = Union[Region, CssSelector, AnyWebElement, BySelector]  # typedef



from __future__ import absolute_import, division, print_function

from enum import Enum
from typing import Any, Text

from ..utils.general_utils import DeprecatedEnumVariant
from .device_name import DeviceName
from .ios_device_name import IosDeviceName


class ScreenOrientation(Enum):
    PORTRAIT = "portrait"
    LANDSCAPE = "landscape"


class IosVersion(Enum):
    LATEST = "latest"
    ONE_VERSION_BACK = "latest-1"


class AndroidVersion(Enum):
    LATEST = "latest"
    ONE_VERSION_BACK = "latest-1"


class VisualGridOption(object):
    def __init__(self, key, value):
        # type: (Text, Any) -> None
        self.key = key
        self.value = value

    def __eq__(self, other):
        # type: (VisualGridOption) -> bool
        return other and self.key == other.key and self.value == other.value



from __future__ import absolute_import, division, print_function

from . import _device_target
from ._android_multi_device_target import *  # noqa
from ._device_target import AndroidDeviceTarget, DeviceTarget, IosDeviceTarget  # noqa
from ._ios_multi_device_target import *  # noqa
from .config import *  # noqa
from .device_name import *  # noqa
from .ios_device_name import *  # noqa
from .render_browser_info import *  # noqa



from __future__ import absolute_import, division, print_function

import marshmallow
from marshmallow import EXCLUDE, Schema, post_dump, post_load, pre_load
from marshmallow.fields import (
    Boolean,
    DateTime,
    Dict,
    Float,
    Function,
    Integer,
    List,
    Nested,
    Raw,
    String,
)
from marshmallow.schema import BaseSchema, SchemaMeta

__all__ = [
    "BaseSchema",
    "Boolean",
    "DateTime",
    "Dict",
    "Raw",
    "Float",
    "Integer",
    "List",
    "Nested",
    "Schema",
    "SchemaMeta",
    "String",
    "Function",
    "post_dump",
    "post_load",
    "pre_load",
    "EXCLUDE",
]



from __future__ import absolute_import, division, print_function

from enum import Enum
from typing import Optional

import attr

from ..utils.general_utils import DeprecatedEnumVariant


class BrowserType(Enum):
    CHROME = "chrome"
    CHROME_ONE_VERSION_BACK = "chrome-one-version-back"
    CHROME_TWO_VERSIONS_BACK = "chrome-two-versions-back"
    FIREFOX = "firefox"
    FIREFOX_ONE_VERSION_BACK = "firefox-one-version-back"
    FIREFOX_TWO_VERSIONS_BACK = "firefox-two-versions-back"
    SAFARI = "safari"
    SAFARI_ONE_VERSION_BACK = "safari-one-version-back"
    SAFARI_TWO_VERSIONS_BACK = "safari-two-versions-back"
    IE_10 = "ie10"
    IE_11 = "ie11"
    EDGE_LEGACY = "edgelegacy"
    EDGE_CHROMIUM = "edgechromium"
    EDGE_CHROMIUM_ONE_VERSION_BACK = "edgechromium-one-version-back"
    EDGE_CHROMIUM_TWO_VERSIONS_BACK = "edgechromium-two-versions-back"

    @DeprecatedEnumVariant
    def EDGE(self):
        # type: () -> BrowserType
        """
        The `EDGE` option that is being used in your browsers configuration
        will soon be deprecated. Please change it to either `EDGE_LEGACY`
        for the legacy version or to `EDGE_CHROMIUM` for the
        new Chromium-based version.
        """
        return self.EDGE_LEGACY


class StitchMode(Enum):
    """
    The type of methods for stitching full-page screenshots.
    """

    Scroll = "Scroll"
    CSS = "CSS"


@attr.s
class MobileOptions(object):
    keep_navigation_bar = attr.ib(default=None)  # type: Optional[bool]



from typing import Optional, Text

import attr

from .config import AndroidVersion, IosVersion, ScreenOrientation


@attr.s
class DeviceTarget(object):
    device_name = attr.ib()  # type: Text
    screen_orientation = attr.ib(default=None)  # type: Optional[ScreenOrientation]

    def landscape(self):
        self.screen_orientation = ScreenOrientation.LANDSCAPE
        return self

    def portrait(self):
        self.screen_orientation = ScreenOrientation.PORTRAIT
        return self


@attr.s
class IosDeviceTarget(DeviceTarget):
    device_name = attr.ib()  # type: Text
    screen_orientation = attr.ib(
        default=ScreenOrientation.PORTRAIT, converter=ScreenOrientation
    )  # type: Optional[ScreenOrientation]
    version = attr.ib(
        default=None, converter=attr.converters.optional(IosVersion)
    )  # type: Optional[IosVersion]


@attr.s
class AndroidDeviceTarget(DeviceTarget):
    device_name = attr.ib()  # type: Text
    screen_orientation = attr.ib(
        default=ScreenOrientation.PORTRAIT, converter=ScreenOrientation
    )  # type: Optional[ScreenOrientation]
    version = attr.ib(
        default=None, converter=attr.converters.optional(AndroidVersion)
    )  # type: Optional[AndroidVersion]



from __future__ import absolute_import, division, print_function

from typing import TYPE_CHECKING

import attr

from applitools.common._dynamic_regions import DynamicSettings

from ..accessibility import AccessibilityRegionType
from ..geometry import AccessibilityRegion, DynamicRegion, Rectangle, Region
from ..match import FloatingBounds

if TYPE_CHECKING:
    from typing import Optional, Text, Union

    from ..utils.custom_types import CodedRegionPadding
    from .target_path import RegionLocator

__all__ = (
    "AccessibilityRegionByRectangle",
    "AccessibilityRegionBySelector",
    "FloatingRegionByRectangle",
    "FloatingRegionBySelector",
    "GetAccessibilityRegion",
    "GetFloatingRegion",
    "GetRegion",
    "RegionByRectangle",
    "RegionBySelector",
    "GetDynamicRegion",
    "DynamicRegionBySelector",
    "DynamicRegionByRectangle",
)


class GetRegion(object):
    pass


@attr.s
class RegionByRectangle(GetRegion):
    _region = attr.ib()  # type: Union[Region, Rectangle]


@attr.s
class RegionBySelector(GetRegion):
    _target_path = attr.ib()  # type: RegionLocator
    padding = attr.ib(default=None)  # type: Optional[CodedRegionPadding]
    region_id = attr.ib(default=None)  # type: Optional[Text]


class GetFloatingRegion(GetRegion):
    pass


@attr.s
class FloatingRegionByRectangle(GetFloatingRegion):
    _rect = attr.ib()  # type: Union[Region, Rectangle]
    _bounds = attr.ib()  # type: FloatingBounds

    @property
    def floating_bounds(self):
        return self._bounds


@attr.s
class FloatingRegionBySelector(GetFloatingRegion):
    _target_path = attr.ib()  # type: RegionLocator
    _bounds = attr.ib()  # type: FloatingBounds


class GetAccessibilityRegion(GetRegion):
    pass


@attr.s
class AccessibilityRegionByRectangle(GetAccessibilityRegion):
    _rect = attr.ib()  # type: Union[Region, Rectangle, AccessibilityRegion]
    _type = attr.ib(default=None)  # type: Optional[AccessibilityRegionType]

    @property
    def accessibility_type(self):
        # type: () -> AccessibilityRegionType
        if self._type:
            return self._type
        return self._rect.type


@attr.s
class AccessibilityRegionBySelector(GetAccessibilityRegion):
    _target_path = attr.ib()  # type: RegionLocator
    _type = attr.ib()  # type: AccessibilityRegionType


class GetDynamicRegion(GetRegion):
    pass


@attr.s
class DynamicRegionByRectangle(GetDynamicRegion):
    _rect = attr.ib()  # type: Union[Region, Rectangle, DynamicRegion]
    _dynamic_settings = attr.ib(default=None)  # type: Optional[DynamicSettings]
    padding = attr.ib(default=None)  # type: Optional[CodedRegionPadding]


@attr.s
class DynamicRegionBySelector(GetDynamicRegion):
    _target_path = attr.ib()  # type: RegionLocator
    _dynamic_settings = attr.ib(default=None)  # type: Optional[DynamicSettings]
    padding = attr.ib(default=None)  # type: Optional[CodedRegionPadding]



try:  # assume python>=3.9
    from importlib.metadata import version as _version
except ImportError:  # older python
    from pkg_resources import get_distribution as _distribution

    def _version(package_name):
        return _distribution(package_name).version


__version__ = _version("eyes-common")



# GENERATED FILE #
from enum import Enum

from applitools.common.utils.general_utils import DeprecatedEnumVariant

__all__ = ("AccessibilityRegionType",)


class AccessibilityRegionType(Enum):
    IgnoreContrast = "IgnoreContrast"
    RegularText = "RegularText"
    LargeText = "LargeText"
    BoldText = "BoldText"
    GraphicalObject = "GraphicalObject"



from __future__ import absolute_import, division, print_function, unicode_literals

from enum import Enum


class Feature(Enum):
    TARGET_WINDOW_CAPTURES_SELECTED_FRAME = "Target window captures selected frame"
    SCALE_MOBILE_APP = "Scale mobile app"



"""Context variable management for marshmallow schema serialization.

This module provides thread-safe context variable management to replace
the deprecated 'context' parameter in marshmallow 3+.
"""

from __future__ import absolute_import, division, print_function

import contextvars
from contextlib import contextmanager
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Generator, Optional

    from .object_registry import ObjectRegistry


# Global context variable for object registry
_object_registry_context = contextvars.ContextVar(
    "object_registry", default=None
)  # type: contextvars.ContextVar[Optional[ObjectRegistry]]


def get_object_registry():
    # type: () -> ObjectRegistry
    """Get the current object registry from context.

    Returns:
        ObjectRegistry: The current object registry instance.

    Raises:
        RuntimeError: If no object registry is set in the current context.
    """
    registry = _object_registry_context.get()
    if registry is None:
        raise RuntimeError(
            "No object registry found in context. "
            "Ensure you're calling this within a set_object_registry() context manager."
        )
    return registry


@contextmanager
def set_object_registry(registry):
    # type: (ObjectRegistry) -> Generator[ObjectRegistry, None, None]
    """Context manager to set the object registry for the current context.

    Args:
        registry: The ObjectRegistry instance to set in the context.

    Yields:
        ObjectRegistry: The registry instance that was set.

    Example:
        with set_object_registry(my_registry) as registry:
            # Now schema fields can access the registry via get_object_registry()
            result = some_schema.dump(data)
    """
    token = _object_registry_context.set(registry)
    try:
        yield registry
    finally:
        _object_registry_context.reset(token)


def has_object_registry():
    # type: () -> bool
    """Check if an object registry is currently set in the context.

    Returns:
        bool: True if an object registry is available, False otherwise.
    """
    return _object_registry_context.get() is not None



from __future__ import absolute_import, division, print_function

import attr

from .. import deprecated
from ._guidelines_version import AccessibilityGuidelinesVersion
from ._level import AccessibilityLevel
from ._region_type import AccessibilityRegionType
from ._status import AccessibilityStatus

__all__ = (
    "AccessibilityGuidelinesVersion",
    "AccessibilityLevel",
    "AccessibilityStatus",
    "AccessibilityRegionType",
    "AccessibilitySettings",
    "SessionAccessibilityStatus",
)


@attr.s(init=False)
class AccessibilitySettings(object):
    level = attr.ib(type=AccessibilityLevel)  # type: AccessibilityLevel
    guidelines_version = attr.ib(
        type=AccessibilityGuidelinesVersion
    )  # type: AccessibilityGuidelinesVersion

    def __init__(
        self,
        level,  # type: AccessibilityLevel
        guidelines_version,  # type: AccessibilityGuidelinesVersion
    ):
        # type: (...) -> None
        self.level = AccessibilityLevel(level)
        self.guidelines_version = AccessibilityGuidelinesVersion(guidelines_version)


@attr.s(init=False)
class SessionAccessibilityStatus(object):
    level = attr.ib(type=AccessibilityLevel)  # type: AccessibilityLevel
    version = attr.ib(
        type=AccessibilityGuidelinesVersion
    )  # type: AccessibilityGuidelinesVersion
    status = attr.ib(type=AccessibilityStatus)  # type: AccessibilityStatus

    def __init__(
        self,
        status,  # type: AccessibilityStatus
        level,  # type: AccessibilityLevel
        version,  # type: AccessibilityGuidelinesVersion
    ):
        self.level = AccessibilityLevel(level)
        self.version = AccessibilityGuidelinesVersion(version)
        self.status = AccessibilityStatus(status)

    @property
    @deprecated.attribute("use `version` instead")
    def guidelines_version(self):
        # type: () -> AccessibilityGuidelinesVersion
        return self.version



from __future__ import absolute_import, division, print_function, unicode_literals

from ..common import deprecated
from ..common.batch_close import BatchClose
from ..common.cut import FixedCutProvider, NullCutProvider, UnscaledFixedCutProvider
from ..common.extract_text import TextRegionSettings
from ..common.feature import Feature
from ..common.fluent import (
    CheckSettings,
    CheckSettingsValues,
    FloatingRegionByRectangle,
    GetFloatingRegion,
    GetRegion,
    RegionByRectangle,
)
from ..common.locators import VisualLocator, VisualLocatorSettings
from ..common.triggers import MouseTrigger, TextTrigger

deprecated.module(__name__)
__all__ = (
    "BatchClose",
    "CheckSettings",
    "CheckSettingsValues",
    "Feature",
    "FixedCutProvider",
    "FloatingRegionByRectangle",
    "GetFloatingRegion",
    "GetRegion",
    "MouseTrigger",
    "NullCutProvider",
    "RegionByRectangle",
    "TextRegionSettings",
    "TextTrigger",
    "UnscaledFixedCutProvider",
    "VisualLocator",
    "VisualLocatorSettings",
    "__version__",
)
